from typing import Annotated

from arcade_tdk import ToolContext, tool
from stripe_agent_toolkit.api import StripeAPI


def run_stripe_tool(context: ToolContext, method_name: str, params: dict) -> str:
    """
    Helper function that retrieves the Stripe secret key, initializes the API,
    and executes the specified method with the provided parameters.
    """
    api_key = context.get_secret("STRIPE_SECRET_KEY")
    stripe_api = StripeAPI(secret_key=api_key, context=None)
    params = {k: v for k, v in params.items() if v is not None}
    return stripe_api.run(method_name, **params)  # type: ignore[no-any-return]


@tool(requires_secrets=["STRIPE_SECRET_KEY"])
def create_customer(
    context: ToolContext,
    name: Annotated[str, "The name of the customer."],
    email: Annotated[str | None, "The email of the customer."] = None,
) -> Annotated[str, "This tool will create a customer in Stripe."]:
    """This tool will create a customer in Stripe."""
    return run_stripe_tool(context, "create_customer", {"name": name, "email": email})


@tool(requires_secrets=["STRIPE_SECRET_KEY"])
def list_customers(
    context: ToolContext,
    limit: Annotated[
        int | None,
        "A limit on the number of objects to be returned. Limit can range between 1 and 100.",
    ] = None,
    email: Annotated[
        str | None,
        "A case-sensitive filter on the list based on the customer's email field. "
        "The value must be a string.",
    ] = None,
) -> Annotated[str, "This tool will fetch a list of Customers from Stripe."]:
    """This tool will fetch a list of Customers from Stripe."""
    return run_stripe_tool(context, "list_customers", {"limit": limit, "email": email})


@tool(requires_secrets=["STRIPE_SECRET_KEY"])
def create_product(
    context: ToolContext,
    name: Annotated[str, "The name of the product."],
    description: Annotated[str | None, "The description of the product."] = None,
) -> Annotated[str, "This tool will create a product in Stripe."]:
    """This tool will create a product in Stripe."""
    return run_stripe_tool(context, "create_product", {"name": name, "description": description})


@tool(requires_secrets=["STRIPE_SECRET_KEY"])
def list_products(
    context: ToolContext,
    limit: Annotated[
        int | None,
        "A limit on the number of objects to be returned. Limit can range between 1 and 100, "
        "and the default is 10.",
    ] = None,
) -> Annotated[str, "This tool will fetch a list of Products from Stripe."]:
    """This tool will fetch a list of Products from Stripe."""
    return run_stripe_tool(context, "list_products", {"limit": limit})


@tool(requires_secrets=["STRIPE_SECRET_KEY"])
def create_price(
    context: ToolContext,
    product: Annotated[str, "The ID of the product to create the price for."],
    unit_amount: Annotated[int, "The unit amount of the price in cents."],
    currency: Annotated[str, "The currency of the price."],
) -> Annotated[str, "This tool will create a price in Stripe. If a product has not already been"]:
    """This tool will create a price in Stripe. If a product has not already been"""
    return run_stripe_tool(
        context,
        "create_price",
        {"product": product, "unit_amount": unit_amount, "currency": currency},
    )


@tool(requires_secrets=["STRIPE_SECRET_KEY"])
def list_prices(
    context: ToolContext,
    product: Annotated[str | None, "The ID of the product to list prices for."] = None,
    limit: Annotated[
        int | None,
        "A limit on the number of objects to be returned. Limit can range between 1 and 100, "
        "and the default is 10.",
    ] = None,
) -> Annotated[str, "This tool will fetch a list of Prices from Stripe."]:
    """This tool will fetch a list of Prices from Stripe."""
    return run_stripe_tool(context, "list_prices", {"product": product, "limit": limit})


@tool(requires_secrets=["STRIPE_SECRET_KEY"])
def create_payment_link(
    context: ToolContext,
    price: Annotated[str, "The ID of the price to create the payment link for."],
    quantity: Annotated[int, "The quantity of the product to include."],
) -> Annotated[str, "This tool will create a payment link in Stripe."]:
    """This tool will create a payment link in Stripe."""
    return run_stripe_tool(context, "create_payment_link", {"price": price, "quantity": quantity})


@tool(requires_secrets=["STRIPE_SECRET_KEY"])
def list_invoices(
    context: ToolContext,
    customer: Annotated[str | None, "The ID of the customer to list invoices for."] = None,
    limit: Annotated[
        int | None,
        "A limit on the number of objects to be returned. Limit can range between 1 and 100, "
        "and the default is 10.",
    ] = None,
) -> Annotated[str, "This tool will list invoices in Stripe."]:
    """This tool will list invoices in Stripe."""
    return run_stripe_tool(context, "list_invoices", {"customer": customer, "limit": limit})


@tool(requires_secrets=["STRIPE_SECRET_KEY"])
def create_invoice(
    context: ToolContext,
    customer: Annotated[str, "The ID of the customer to create the invoice for."],
    days_until_due: Annotated[int | None, "The number of days until the invoice is due."] = None,
) -> Annotated[str, "This tool will create an invoice in Stripe."]:
    """This tool will create an invoice in Stripe."""
    return run_stripe_tool(
        context, "create_invoice", {"customer": customer, "days_until_due": days_until_due}
    )


@tool(requires_secrets=["STRIPE_SECRET_KEY"])
def create_invoice_item(
    context: ToolContext,
    customer: Annotated[str, "The ID of the customer to create the invoice item for."],
    price: Annotated[str, "The ID of the price for the item."],
    invoice: Annotated[str, "The ID of the invoice to create the item for."],
) -> Annotated[str, "This tool will create an invoice item in Stripe."]:
    """This tool will create an invoice item in Stripe."""
    return run_stripe_tool(
        context, "create_invoice_item", {"customer": customer, "price": price, "invoice": invoice}
    )


@tool(requires_secrets=["STRIPE_SECRET_KEY"])
def finalize_invoice(
    context: ToolContext, invoice: Annotated[str, "The ID of the invoice to finalize."]
) -> Annotated[str, "This tool will finalize an invoice in Stripe."]:
    """This tool will finalize an invoice in Stripe."""
    return run_stripe_tool(context, "finalize_invoice", {"invoice": invoice})


@tool(requires_secrets=["STRIPE_SECRET_KEY"])
def retrieve_balance(
    context: ToolContext,
) -> Annotated[str, "This tool will retrieve the balance from Stripe. It takes no input."]:
    """This tool will retrieve the balance from Stripe. It takes no input."""
    return run_stripe_tool(context, "retrieve_balance", {})


@tool(requires_secrets=["STRIPE_SECRET_KEY"])
def create_refund(
    context: ToolContext,
    payment_intent: Annotated[str, "The ID of the PaymentIntent to refund."],
    amount: Annotated[int | None, "The amount to refund in cents."] = None,
) -> Annotated[str, "This tool will refund a payment intent in Stripe."]:
    """This tool will refund a payment intent in Stripe."""
    return run_stripe_tool(
        context, "create_refund", {"payment_intent": payment_intent, "amount": amount}
    )


@tool(requires_secrets=["STRIPE_SECRET_KEY"])
def list_payment_intents(
    context: ToolContext,
    customer: Annotated[str | None, "The ID of the customer to list payment intents for."] = None,
    limit: Annotated[
        int | None,
        "A limit on the number of objects to be returned. Limit can range between 1 and 100.",
    ] = None,
) -> Annotated[str, "This tool will list payment intents in Stripe."]:
    """This tool will list payment intents in Stripe."""
    return run_stripe_tool(context, "list_payment_intents", {"customer": customer, "limit": limit})


@tool(requires_secrets=["STRIPE_SECRET_KEY"])
def create_billing_portal_session(
    context: ToolContext,
    customer: Annotated[str, "The ID of the customer to create the billing portal session for."],
    return_url: Annotated[str | None, "The default URL to return to afterwards."] = None,
) -> Annotated[str, "This tool will create a billing portal session."]:
    """This tool will create a billing portal session."""
    return run_stripe_tool(
        context, "create_billing_portal_session", {"customer": customer, "return_url": return_url}
    )



from enum import Enum


class GoogleFlightsTravelClass(Enum):
    ECONOMY = "ECONOMY"
    PREMIUM_ECONOMY = "PREMIUM_ECONOMY"
    BUSINESS = "BUSINESS"
    FIRST = "FIRST"

    def to_api_value(self) -> int:
        _map = {
            "ECONOMY": 1,
            "PREMIUM_ECONOMY": 2,
            "BUSINESS": 3,
            "FIRST": 4,
        }
        return _map[self.value]


class GoogleFlightsMaxStops(Enum):
    ANY = "ANY"
    NONSTOP = "NONSTOP"
    ONE = "ONE"
    TWO = "TWO"

    def to_api_value(self) -> int:
        _map = {
            "ANY": 0,
            "NONSTOP": 1,
            "ONE": 2,
            "TWO": 3,
        }
        return _map[self.value]


class GoogleFlightsSortBy(Enum):
    TOP_FLIGHTS = "TOP_FLIGHTS"
    PRICE = "PRICE"
    DEPARTURE_TIME = "DEPARTURE_TIME"
    ARRIVAL_TIME = "ARRIVAL_TIME"
    DURATION = "DURATION"
    EMISSIONS = "EMISSIONS"

    def to_api_value(self) -> int:
        _map = {
            "TOP_FLIGHTS": 1,
            "PRICE": 2,
            "DEPARTURE_TIME": 3,
            "ARRIVAL_TIME": 4,
            "DURATION": 5,
            "EMISSIONS": 6,
        }
        return _map[self.value]



LANGUAGE_CODES = {
    "ar": "Arabic",
    "bn": "Bengali",
    "da": "Danish",
    "de": "German",
    "el": "Greek",
    "en": "English",
    "es": "Spanish",
    "fi": "Finnish",
    "fr": "French",
    "hi": "Hindi",
    "hu": "Hungarian",
    "id": "Indonesian",
    "it": "Italian",
    "ja": "Japanese",
    "ko": "Korean",
    "nl": "Dutch",
    "ms": "Malay",
    "no": "Norwegian",
    "pcm": "Nigerian Pidgin",
    "pl": "Polish",
    "pt": "Portuguese",
    "pt-br": "Portuguese (Brazil)",
    "pt-pt": "Portuguese (Portugal)",
    "ru": "Russian",
    "sv": "Swedish",
    "tl": "Filipino",
    "tr": "Turkish",
    "uk": "Ukrainian",
    "zh": "Chinese",
    "zh-cn": "Chinese (Simplified)",
    "zh-tw": "Chinese (Traditional)",
}



import os
import warnings
from collections.abc import Iterator
from typing import Any, Optional, Union

from arcadepy import NOT_GIVEN, Arcade, AsyncArcade
from arcadepy.types import ToolDefinition
from arcadepy.types.shared import AuthorizationResponse
from langchain_core.tools import StructuredTool

from langchain_arcade._utilities import wrap_arcade_tool

ClientType = Union[Arcade, AsyncArcade]


class LangChainToolManager:
    """
    Base tool manager for LangChain framework.
    Provides a common interface for both synchronous and asynchronous tool managers.

    This class handles the storage and retrieval of tool definitions and provides
    common functionality used by both synchronous and asynchronous implementations.
    """

    def __init__(self) -> None:
        self._tools: dict[str, ToolDefinition] = {}

    @property
    def tools(self) -> list[str]:
        """
        Get the list of tools by name in the manager.

        Returns:
            A list of tool names (strings) currently stored in the manager.
        """
        return list(self._tools.keys())

    def __len__(self) -> int:
        """Return the number of tools in the manager."""
        return len(self._tools)

    def _get_client_config(self, **kwargs: Any) -> dict[str, Any]:
        """
        Get the client configurations from environment variables and kwargs.

        If api_key or base_url are in the kwargs, they will be used.
        Otherwise, the environment variables ARCADE_API_KEY and ARCADE_BASE_URL will be used.
        If both are provided, the kwargs will take precedence.

        Args:
            **kwargs: Keyword arguments that may contain api_key and base_url.

        Returns:
            A dictionary of client configuration parameters.
        """
        client_kwargs = {
            "api_key": kwargs.get("api_key", os.getenv("ARCADE_API_KEY")),
        }
        base_url = kwargs.get("base_url", os.getenv("ARCADE_BASE_URL"))
        if base_url:
            client_kwargs["base_url"] = base_url
        return client_kwargs

    def _get_tool_definition(self, tool_name: str) -> ToolDefinition:
        """
        Get a tool definition by name, raising an error if not found.

        Args:
            tool_name: The name of the tool to retrieve.

        Returns:
            The ToolDefinition for the specified tool.

        Raises:
            ValueError: If the tool is not found in the manager.
        """
        try:
            return self._tools[tool_name]
        except KeyError:
            raise ValueError(f"Tool '{tool_name}' not found in this manager instance")

    def __getitem__(self, tool_name: str) -> ToolDefinition:
        """
        Get a tool definition by name using dictionary-like access.

        Args:
            tool_name: The name of the tool to retrieve.

        Returns:
            The ToolDefinition for the specified tool.

        Raises:
            ValueError: If the tool is not found in the manager.
        """
        return self._get_tool_definition(tool_name)

    def requires_auth(self, tool_name: str) -> bool:
        """
        Check if a tool requires authorization.

        Args:
            tool_name: The name of the tool to check.

        Returns:
            True if the tool requires authorization, False otherwise.
        """
        tool_def = self._get_tool_definition(tool_name)
        if tool_def.requirements is None:
            return False
        return tool_def.requirements.authorization is not None


class ToolManager(LangChainToolManager):
    """
    Synchronous Arcade tool manager for LangChain framework.

    This class wraps Arcade tools as LangChain StructuredTool objects for integration
    with synchronous operations.

    Example:
        >>> manager = ToolManager(api_key="your-api-key")
        >>> # Initialize with specific tools and toolkits
        >>> manager.init_tools(tools=["Search.SearchGoogle"], toolkits=["Weather"])
        >>> # Get tools as LangChain StructuredTools
        >>> langchain_tools = manager.to_langchain()
        >>> # Handle authorization for tools that require it
        >>> if manager.requires_auth("Search.SearchGoogle"):
        >>>     auth_response = manager.authorize("Search.SearchGoogle", "user_123")
        >>>     manager.wait_for_auth(auth_response.id)
    """

    def __init__(self, client: Optional[Arcade] = None, **kwargs: Any) -> None:
        """
        Initialize the ToolManager.

        Example:
            >>> manager = ToolManager(api_key="your-api-key")
            >>> # or with an existing client
            >>> client = Arcade(api_key="your-api-key")
            >>> manager = ToolManager(client=client)

        Args:
            client: Optional Arcade client instance. If not provided, one will be created.
            **kwargs: Additional keyword arguments to pass to the Arcade client if creating one.
                      Common options include api_key and base_url.
        """
        super().__init__()
        if client is None:
            client_kwargs = self._get_client_config(**kwargs)
            client = Arcade(**client_kwargs)
        self._client = client

    @property
    def definitions(self) -> list[ToolDefinition]:
        """
        Get the list of tool definitions in the manager.

        Returns:
            A list of ToolDefinition objects currently stored in the manager.
        """
        return list(self._tools.values())

    def __iter__(self) -> Iterator[tuple[str, ToolDefinition]]:
        """
        Iterate over the tools in the manager as (name, definition) pairs.

        Returns:
            Iterator over (tool_name, tool_definition) tuples.
        """
        yield from self._tools.items()

    def to_langchain(
        self, use_interrupts: bool = True, use_underscores: bool = True
    ) -> list[StructuredTool]:
        """
        Get the tools in the manager as LangChain StructuredTool objects.

        Args:
            use_interrupts: Whether to use interrupts for the tool. This is useful
                           for LangGraph workflows where you need to handle tool
                           authorization through state transitions.
            use_underscores: Whether to use underscores for the tool name instead of periods.
                            For example, "Search_SearchGoogle" vs "Search.SearchGoogle".
                            Some model providers like OpenAI work better with underscores.

        Returns:
            List of StructuredTool instances ready to use with LangChain.
        """
        tool_map = _create_tool_map(self.definitions, use_underscores=use_underscores)
        return [
            wrap_arcade_tool(self._client, tool_name, definition, langgraph=use_interrupts)
            for tool_name, definition in tool_map.items()
        ]

    def init_tools(
        self,
        tools: Optional[list[str]] = None,
        toolkits: Optional[list[str]] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        raise_on_empty: bool = True,
    ) -> list[StructuredTool]:
        """
        Initialize the tools in the manager and return them as LangChain tools.

        This will clear any existing tools in the manager and replace them with
        the new tools specified by the tools and toolkits parameters.

        Note: In version 2.0+, this method returns a list of StructuredTool objects.
        In earlier versions, it returned None.

        Example:
            >>> manager = ToolManager(api_key="your-api-key")
            >>> langchain_tools = manager.init_tools(tools=["Search.SearchGoogle"])
            >>> # Use these tools with a LangChain chain or agent
            >>> agent = Agent(tools=langchain_tools, llm=llm)

        Args:
            tools: Optional list of specific tool names to include (e.g., "Search.SearchGoogle").
            toolkits: Optional list of toolkit names to include all tools from (e.g., "Search").
            limit: Optional limit on the number of tools to retrieve per request.
            offset: Optional offset for paginated requests.
            raise_on_empty: Whether to raise an error if no tools or toolkits are provided.

        Returns:
            List of StructuredTool instances ready to use with LangChain.

        Raises:
            ValueError: If no tools or toolkits are provided and raise_on_empty is True.
        """
        tools_list = self._retrieve_tool_definitions(tools, toolkits, raise_on_empty, limit, offset)
        self._tools = _create_tool_map(tools_list)
        return self.to_langchain()

    def authorize(self, tool_name: str, user_id: str) -> AuthorizationResponse:
        """
        Authorize a user for a specific tool.

        Example:
            >>> manager = ToolManager(api_key="your-api-key")
            >>> manager.init_tools(tools=["Gmail.SendEmail"])
            >>> auth_response = manager.authorize("Gmail.SendEmail", "user_123")
            >>> # auth_response.auth_url contains the URL for the user to authorize

        Args:
            tool_name: The name of the tool to authorize.
            user_id: The user ID to authorize. This should be a unique identifier for the user.

        Returns:
            AuthorizationResponse containing authorization details, including the auth_url
            that should be presented to the user to complete authorization.
        """
        return self._client.tools.authorize(tool_name=tool_name, user_id=user_id)

    def is_authorized(self, authorization_id: str) -> bool:
        """
        Check if a tool authorization is complete.

        Example:
            >>> manager = ToolManager(api_key="your-api-key")
            >>> auth_response = manager.authorize("Gmail.SendEmail", "user_123")
            >>> # After user completes authorization
            >>> is_complete = manager.is_authorized(auth_response.id)

        Args:
            authorization_id: The authorization ID to check. This can be the full AuthorizationResponse
                             object or just the ID string.

        Returns:
            True if the authorization is completed, False otherwise.
        """
        # Handle case where entire AuthorizationResponse object is passed
        if hasattr(authorization_id, "id"):
            authorization_id = authorization_id.id

        response = self._client.auth.status(id=authorization_id)
        if response:
            return response.status == "completed"
        return False

    def wait_for_auth(self, authorization_id: str) -> AuthorizationResponse:
        """
        Wait for a tool authorization to complete. This method blocks until
        the authorization is complete or fails.

        Example:
            >>> manager = ToolManager(api_key="your-api-key")
            >>> auth_response = manager.authorize("Gmail.SendEmail", "user_123")
            >>> # Share auth_response.auth_url with the user
            >>> # Wait for the user to complete authorization
            >>> completed_auth = manager.wait_for_auth(auth_response.id)

        Args:
            authorization_id: The authorization ID to wait for. This can be the full
                             AuthorizationResponse object or just the ID string.

        Returns:
            AuthorizationResponse with the completed authorization details.
        """
        # Handle case where entire AuthorizationResponse object is passed
        if hasattr(authorization_id, "id"):
            authorization_id = authorization_id.id

        return self._client.auth.wait_for_completion(authorization_id)

    def _retrieve_tool_definitions(
        self,
        tools: Optional[list[str]] = None,
        toolkits: Optional[list[str]] = None,
        raise_on_empty: bool = True,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> list[ToolDefinition]:
        """
        Retrieve tool definitions from the Arcade client, accounting for pagination.

        Args:
            tools: Optional list of specific tool names to include.
            toolkits: Optional list of toolkit names to include all tools from.
            raise_on_empty: Whether to raise an error if no tools or toolkits are provided.
            limit: Optional limit on the number of tools to retrieve per request.
            offset: Optional offset for paginated requests.

        Returns:
            List of ToolDefinition instances.

        Raises:
            ValueError: If no tools or toolkits are provided and raise_on_empty is True.
        """
        all_tools: list[ToolDefinition] = []

        # If no specific tools or toolkits are requested, raise an error.
        if not tools and not toolkits:
            if raise_on_empty:
                raise ValueError("No tools or toolkits provided to retrieve tool definitions.")
            return []

        # Retrieve individual tools if specified
        if tools:
            for tool_id in tools:
                single_tool = self._client.tools.get(name=tool_id)
                all_tools.append(single_tool)

        # Retrieve tools from specified toolkits
        if toolkits:
            for tk in toolkits:
                # Convert None to NOT_GIVEN for Stainless client
                paginated_tools = self._client.tools.list(
                    toolkit=tk,
                    limit=limit if limit is not None else NOT_GIVEN,
                    offset=offset if offset is not None else NOT_GIVEN,
                )
                all_tools.extend(paginated_tools)

        return all_tools

    def add_tool(self, tool_name: str) -> None:
        """
        Add a single tool to the manager by name.

        Unlike init_tools(), this method preserves existing tools in the manager
        and only adds the specified tool.

        Example:
            >>> manager = ToolManager(api_key="your-api-key")
            >>> manager.add_tool("Gmail.SendEmail")
            >>> manager.add_tool("Search.SearchGoogle")
            >>> # Get all tools including newly added ones
            >>> all_tools = manager.to_langchain()

        Args:
            tool_name: The fully qualified name of the tool to add (e.g., "Search.SearchGoogle")

        Raises:
            ValueError: If the tool cannot be found
        """
        tool = self._client.tools.get(name=tool_name)
        self._tools.update(_create_tool_map([tool]))

    def add_toolkit(
        self, toolkit_name: str, limit: Optional[int] = None, offset: Optional[int] = None
    ) -> None:
        """
        Add all tools from a specific toolkit to the manager.

        Unlike init_tools(), this method preserves existing tools in the manager
        and only adds the tools from the specified toolkit.

        Example:
            >>> manager = ToolManager(api_key="your-api-key")
            >>> manager.add_toolkit("Gmail")
            >>> manager.add_toolkit("Search")
            >>> # Get all tools including newly added ones
            >>> all_tools = manager.to_langchain()

        Args:
            toolkit_name: The name of the toolkit to add (e.g., "Search")
            limit: Optional limit on the number of tools to retrieve per request
            offset: Optional offset for paginated requests

        Raises:
            ValueError: If the toolkit cannot be found or has no tools
        """
        tools = self._client.tools.list(
            toolkit=toolkit_name,
            limit=NOT_GIVEN if limit is None else limit,
            offset=NOT_GIVEN if offset is None else offset,
        )

        for tool in tools:
            self._tools.update(_create_tool_map([tool]))

    def get_tools(
        self,
        tools: Optional[list[str]] = None,
        toolkits: Optional[list[str]] = None,
        langgraph: bool = True,
    ) -> list[StructuredTool]:
        """
        DEPRECATED: Return the tools in the manager as LangChain StructuredTool objects.

        This method is deprecated and will be removed in a future major version.
        Please use `init_tools()` to initialize tools and `to_langchain()` to convert them.

        Args:
            tools: Optional list of tool names to include.
            toolkits: Optional list of toolkits to include.
            langgraph: Whether to use LangGraph-specific behavior
                such as NodeInterrupts for auth.

        Returns:
            List of StructuredTool instances.
        """
        warnings.warn(
            "get_tools() is deprecated and will be removed in the next major version. "
            "Please use init_tools() to initialize tools and to_langchain() to convert them.",
            DeprecationWarning,
            stacklevel=2,
        )

        # Support existing usage pattern
        if tools or toolkits:
            self.init_tools(tools=tools, toolkits=toolkits)

        return self.to_langchain(use_interrupts=langgraph)


class ArcadeToolManager(ToolManager):
    """
    Deprecated alias for ToolManager.

    ArcadeToolManager is deprecated and will be removed in the next major version.
    Please use ToolManager instead.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        warnings.warn(
            "ArcadeToolManager is deprecated and will be removed in the next major version. "
            "Please use ToolManager instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(*args, **kwargs)


class AsyncToolManager(LangChainToolManager):
    """
    Async version of Arcade tool manager for LangChain framework.

    This class wraps Arcade tools as LangChain StructuredTool objects for integration
    with asynchronous operations.

    Example:
        >>> manager = AsyncToolManager(api_key="your-api-key")
        >>> # Initialize with specific tools and toolkits
        >>> await manager.init_tools(tools=["Search.SearchGoogle"], toolkits=["Weather"])
        >>> # Get tools as LangChain StructuredTools
        >>> langchain_tools = await manager.to_langchain()
        >>> # Handle authorization for tools that require it
        >>> if manager.requires_auth("Search.SearchGoogle"):
        >>>     auth_response = await manager.authorize("Search.SearchGoogle", "user_123")
        >>>     await manager.wait_for_auth(auth_response.id)
    """

    def __init__(
        self,
        client: Optional[AsyncArcade] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize the AsyncToolManager.

        Example:
            >>> manager = AsyncToolManager(api_key="your-api-key")
            >>> # or with an existing client
            >>> client = AsyncArcade(api_key="your-api-key")
            >>> manager = AsyncToolManager(client=client)

        Args:
            client: Optional AsyncArcade client instance. If not provided, one will be created.
            **kwargs: Additional keyword arguments to pass to the AsyncArcade client if creating one.
                      Common options include api_key and base_url.
        """
        super().__init__()
        if not client:
            client_kwargs = self._get_client_config(**kwargs)
            client = AsyncArcade(**client_kwargs)
        self._client = client

    @property
    def definitions(self) -> list[ToolDefinition]:
        """
        Get the list of tool definitions in the manager.

        Returns:
            A list of ToolDefinition objects currently stored in the manager.
        """
        return list(self._tools.values())

    def __iter__(self) -> Iterator[tuple[str, ToolDefinition]]:
        """
        Iterate over the tools in the manager as (name, definition) pairs.

        Returns:
            Iterator over (tool_name, tool_definition) tuples.
        """
        yield from self._tools.items()

    async def init_tools(
        self,
        tools: Optional[list[str]] = None,
        toolkits: Optional[list[str]] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        raise_on_empty: bool = True,
    ) -> list[StructuredTool]:
        """
        Initialize the tools in the manager asynchronously and return them as LangChain tools.

        This will clear any existing tools in the manager and replace them with
        the new tools specified by the tools and toolkits parameters.

        Example:
            >>> manager = AsyncToolManager(api_key="your-api-key")
            >>> langchain_tools = await manager.init_tools(tools=["Search.SearchGoogle"])
            >>> # Use these tools with a LangChain chain or agent
            >>> agent = Agent(tools=langchain_tools, llm=llm)

        Args:
            tools: Optional list of specific tool names to include (e.g., "Search.SearchGoogle").
            toolkits: Optional list of toolkit names to include all tools from (e.g., "Search").
            limit: Optional limit on the number of tools to retrieve per request.
            offset: Optional offset for paginated requests.
            raise_on_empty: Whether to raise an error if no tools or toolkits are provided.

        Returns:
            List of StructuredTool instances ready to use with LangChain.

        Raises:
            ValueError: If no tools or toolkits are provided and raise_on_empty is True.
        """
        tools_list = await self._retrieve_tool_definitions(
            tools, toolkits, raise_on_empty, limit, offset
        )
        self._tools.update(_create_tool_map(tools_list))
        return await self.to_langchain()

    async def to_langchain(
        self, use_interrupts: bool = True, use_underscores: bool = True
    ) -> list[StructuredTool]:
        """
        Get the tools in the manager as LangChain StructuredTool objects asynchronously.

        Args:
            use_interrupts: Whether to use interrupts for the tool. This is useful
                           for LangGraph workflows where you need to handle tool
                           authorization through state transitions.
            use_underscores: Whether to use underscores for the tool name instead of periods.
                            For example, "Search_SearchGoogle" vs "Search.SearchGoogle".
                            Some model providers like OpenAI work better with underscores.

        Returns:
            List of StructuredTool instances ready to use with LangChain.
        """
        tool_map = _create_tool_map(self.definitions, use_underscores=use_underscores)
        return [
            wrap_arcade_tool(self._client, tool_name, definition, langgraph=use_interrupts)
            for tool_name, definition in tool_map.items()
        ]

    async def authorize(self, tool_name: str, user_id: str) -> AuthorizationResponse:
        """
        Authorize a user for a tool.

        Example:
            >>> manager = AsyncToolManager(api_key="your-api-key")
            >>> await manager.init_tools(tools=["Gmail.SendEmail"])
            >>> auth_response = await manager.authorize("Gmail.SendEmail", "user_123")
            >>> # auth_response.auth_url contains the URL for the user to authorize

        Args:
            tool_name: The name of the tool to authorize.
            user_id: The user ID to authorize. This should be a unique identifier for the user.

        Returns:
            AuthorizationResponse containing authorization details, including the auth_url
            that should be presented to the user to complete authorization.
        """
        return await self._client.tools.authorize(tool_name=tool_name, user_id=user_id)

    async def is_authorized(self, authorization_id: str) -> bool:
        """
        Check if a tool authorization is complete.

        Example:
            >>> manager = AsyncToolManager(api_key="your-api-key")
            >>> auth_response = await manager.authorize("Gmail.SendEmail", "user_123")
            >>> # After user completes authorization
            >>> is_complete = await manager.is_authorized(auth_response.id)

        Args:
            authorization_id: The authorization ID to check. This can be the full AuthorizationResponse
                             object or just the ID string.

        Returns:
            True if the authorization is completed, False otherwise.
        """
        # Handle case where entire AuthorizationResponse object is passed
        if hasattr(authorization_id, "id"):
            authorization_id = authorization_id.id

        auth_status = await self._client.auth.status(id=authorization_id)
        return auth_status.status == "completed"

    async def wait_for_auth(self, authorization_id: str) -> AuthorizationResponse:
        """
        Wait for a tool authorization to complete. This method blocks until
        the authorization is complete or fails.

        Example:
            >>> manager = AsyncToolManager(api_key="your-api-key")
            >>> auth_response = await manager.authorize("Gmail.SendEmail", "user_123")
            >>> # Share auth_response.auth_url with the user
            >>> # Wait for the user to complete authorization
            >>> completed_auth = await manager.wait_for_auth(auth_response.id)

        Args:
            authorization_id: The authorization ID to wait for. This can be the full
                             AuthorizationResponse object or just the ID string.

        Returns:
            AuthorizationResponse with the completed authorization details.
        """
        # Handle case where entire AuthorizationResponse object is passed
        if hasattr(authorization_id, "id"):
            authorization_id = authorization_id.id

        return await self._client.auth.wait_for_completion(authorization_id)

    async def _retrieve_tool_definitions(
        self,
        tools: Optional[list[str]] = None,
        toolkits: Optional[list[str]] = None,
        raise_on_empty: bool = True,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> list[ToolDefinition]:
        """
        Retrieve tool definitions asynchronously from the Arcade client, accounting for pagination.

        Args:
            tools: Optional list of specific tool names to include.
            toolkits: Optional list of toolkit names to include all tools from.
            raise_on_empty: Whether to raise an error if no tools or toolkits are provided.
            limit: Optional limit on the number of tools to retrieve per request.
            offset: Optional offset for paginated requests.

        Returns:
            List of ToolDefinition instances.

        Raises:
            ValueError: If no tools or toolkits are provided and raise_on_empty is True.
        """
        all_tools: list[ToolDefinition] = []

        # If no specific tools or toolkits are requested, raise an error.
        if not tools and not toolkits:
            if raise_on_empty:
                raise ValueError("No tools or toolkits provided to retrieve tool definitions.")
            return []

        # First, gather single tools if the user specifically requested them.
        if tools:
            for tool_id in tools:
                # ToolsResource.get(...) returns a single ToolDefinition.
                single_tool = await self._client.tools.get(name=tool_id)
                all_tools.append(single_tool)

        # Next, gather tool definitions from any requested toolkits.
        if toolkits:
            for tk in toolkits:
                # Convert None to NOT_GIVEN for Stainless client
                paginated_tools = await self._client.tools.list(
                    toolkit=tk,
                    limit=NOT_GIVEN if limit is None else limit,
                    offset=NOT_GIVEN if offset is None else offset,
                )
                async for tool in paginated_tools:
                    all_tools.append(tool)

        return all_tools

    async def add_tool(self, tool_name: str) -> None:
        """
        Add a single tool to the manager by name.

        Unlike init_tools(), this method preserves existing tools in the manager
        and only adds the specified tool.

        Example:
            >>> manager = AsyncToolManager(api_key="your-api-key")
            >>> await manager.add_tool("Gmail.SendEmail")
            >>> await manager.add_tool("Search.SearchGoogle")
            >>> # Get all tools including newly added ones
            >>> all_tools = await manager.to_langchain()

        Args:
            tool_name: The fully qualified name of the tool to add (e.g., "Search.SearchGoogle")

        Raises:
            ValueError: If the tool cannot be found
        """
        tool = await self._client.tools.get(name=tool_name)
        self._tools.update(_create_tool_map([tool]))

    async def add_toolkit(
        self, toolkit_name: str, limit: Optional[int] = None, offset: Optional[int] = None
    ) -> None:
        """
        Add all tools from a specific toolkit to the manager.

        Unlike init_tools(), this method preserves existing tools in the manager
        and only adds the tools from the specified toolkit.

        Example:
            >>> manager = AsyncToolManager(api_key="your-api-key")
            >>> await manager.add_toolkit("Gmail")
            >>> await manager.add_toolkit("Search")
            >>> # Get all tools including newly added ones
            >>> all_tools = await manager.to_langchain()

        Args:
            toolkit_name: The name of the toolkit to add (e.g., "Search")
            limit: Optional limit on the number of tools to retrieve per request
            offset: Optional offset for paginated requests

        Raises:
            ValueError: If the toolkit cannot be found or has no tools
        """
        paginated_tools = await self._client.tools.list(
            toolkit=toolkit_name,
            limit=NOT_GIVEN if limit is None else limit,
            offset=NOT_GIVEN if offset is None else offset,
        )

        async for tool in paginated_tools:
            self._tools.update(_create_tool_map([tool]))

    async def get_tools(
        self,
        tools: Optional[list[str]] = None,
        toolkits: Optional[list[str]] = None,
        langgraph: bool = True,
    ) -> list[StructuredTool]:
        """
        DEPRECATED: Return the tools in the manager as LangChain StructuredTool objects.

        This method is deprecated and will be removed in a future major version.
        Please use `init_tools()` to initialize tools and `to_langchain()` to convert them.

        Args:
            tools: Optional list of tool names to include.
            toolkits: Optional list of toolkits to include.
            langgraph: Whether to use LangGraph-specific behavior
                such as NodeInterrupts for auth.

        Returns:
            List of StructuredTool instances.
        """
        warnings.warn(
            "get_tools() is deprecated and will be removed in the next major version. "
            "Please use init_tools() to initialize tools and to_langchain() to convert them.",
            DeprecationWarning,
            stacklevel=2,
        )

        # Support existing usage pattern
        if tools or toolkits:
            return await self.init_tools(tools=tools, toolkits=toolkits)
        return []


def _create_tool_map(
    tools: list[ToolDefinition],
    use_underscores: bool = True,
) -> dict[str, ToolDefinition]:
    """
    Build a dictionary that maps the "full_tool_name" to the tool definition.

    Args:
        tools: List of ToolDefinition objects to map.
        use_underscores: Whether to use underscores instead of periods in tool names.
                         For example, "Search_SearchGoogle" vs "Search.SearchGoogle".

    Returns:
        Dictionary mapping tool names to tool definitions.

    Note:
        This is a temporary solution to support the naming convention of certain model providers
        like OpenAI, which work better with underscores in tool names.
    """
    tool_map: dict[str, ToolDefinition] = {}
    for tool in tools:
        # Ensure toolkit name and tool name are not None before creating the key
        toolkit_name = tool.toolkit.name if tool.toolkit and tool.toolkit.name else None
        if toolkit_name and tool.name:
            if use_underscores:
                tool_name = f"{toolkit_name}_{tool.name}"
            else:
                tool_name = f"{toolkit_name}.{tool.name}"
            tool_map[tool_name] = tool
    return tool_map



import json
import logging
from enum import Enum
from typing import Any

from arcade_core.catalog import MaterializedTool

# Type aliases for MCP types
MCPTool = dict[str, Any]
MCPTextContent = dict[str, Any]
MCPImageContent = dict[str, Any]
MCPEmbeddedResource = dict[str, Any]
MCPContent = MCPTextContent | MCPImageContent | MCPEmbeddedResource

logger = logging.getLogger("arcade.mcp")


def create_mcp_tool(tool: MaterializedTool) -> dict[str, Any] | None:  # noqa: C901
    """
    Create an MCP-compatible tool definition from an Arcade tool.

    Args:
        tool: An Arcade tool object

    Returns:
        An MCP tool definition or None if the tool cannot be converted
    """
    try:
        name = getattr(tool.definition, "fully_qualified_name", None) or getattr(
            tool.definition, "name", "unknown"
        )
        description = getattr(tool.definition, "description", "No description available")

        # Extract parameters from the input model
        parameters = {}
        required = []

        if (
            hasattr(tool, "input_model")
            and tool.input_model is not None
            and hasattr(tool.input_model, "model_fields")
        ):
            for field_name, field in tool.input_model.model_fields.items():
                # Skip internal tool context parameters
                if field_name == getattr(
                    tool.definition.input, "tool_context_parameter_name", None
                ):
                    continue

                # Get field type information
                field_type = getattr(field, "annotation", None)
                field_type_name = "string"  # default

                # Safety check for field_type
                if field_type is int:
                    field_type_name = "integer"
                elif field_type is float:
                    field_type_name = "number"
                elif field_type is bool:
                    field_type_name = "boolean"
                elif field_type is list or str(field_type).startswith("list["):
                    field_type_name = "array"
                elif field_type is dict or str(field_type).startswith("dict["):
                    field_type_name = "object"

                # Get description with fallback
                field_description = getattr(field, "description", None)
                if not field_description:
                    field_description = f"Parameter: {field_name}"

                # Create parameter definition
                param_def = {
                    "type": field_type_name,
                    "description": field_description,
                }

                # Enum support: if the field annotation is an Enum, add allowed values
                enum_type = None
                if hasattr(field, "annotation"):
                    ann = field.annotation
                    # Handle typing.Annotated[Enum, ...]
                    if getattr(ann, "__origin__", None) is not None and hasattr(ann, "__args__"):
                        for arg in ann.__args__:  # type: ignore[union-attr]
                            if isinstance(arg, type) and issubclass(arg, Enum):
                                enum_type = arg
                                break
                    elif isinstance(ann, type) and issubclass(ann, Enum):
                        enum_type = ann
                if enum_type is not None:
                    param_def["enum"] = [e.value for e in enum_type]

                parameters[field_name] = param_def

                # In Pydantic v2, check if field is required based on default value
                try:
                    if field.is_required():
                        required.append(field_name)
                except (AttributeError, TypeError):
                    # Fallback if is_required() doesn't exist or fails
                    try:
                        has_default = getattr(field, "default", None) is not None
                        has_factory = getattr(field, "default_factory", None) is not None
                        if not (has_default or has_factory):
                            required.append(field_name)
                    except Exception:
                        # Ultimate fallback - assume required if we can't determine
                        logger.debug(
                            f"Could not determine if field {field_name} is required, assuming optional"
                        )

        # Create the input schema with explicit properties and required fields
        input_schema = {
            "type": "object",
            "properties": parameters,
        }

        # Only include required field if we have required parameters
        if required:
            input_schema["required"] = required

        # Add annotations based on tool metadata
        annotations = {}

        # Use tool name as title if available
        annotations["title"] = getattr(tool.definition, "title", str(name).replace(".", "_"))

        # Determine hints based on tool properties
        if hasattr(tool.definition, "metadata"):
            metadata = tool.definition.metadata or {}
            annotations["readOnlyHint"] = metadata.get("read_only", False)
            annotations["destructiveHint"] = metadata.get("destructive", False)
            annotations["idempotentHint"] = metadata.get("idempotent", True)
            annotations["openWorldHint"] = metadata.get("open_world", False)

        # Create the final tool definition
        tool_def: MCPTool = {
            "name": str(name).replace(".", "_"),
            "description": str(description),
            "inputSchema": input_schema,
            "annotations": annotations,
        }

        logger.debug(f"Created tool definition for {name}")

    except Exception:
        logger.exception(
            f"Error creating MCP tool definition for {getattr(tool, 'name', str(tool))}"
        )
        return None
    return tool_def


def convert_to_mcp_content(value: Any) -> list[dict[str, Any]]:
    """
    Convert a Python value to MCP-compatible content.
    """
    if value is None:
        return []

    if isinstance(value, (str, bool, int, float)):
        return [{"type": "text", "text": str(value)}]

    if isinstance(value, (dict, list)):
        return [{"type": "text", "text": json.dumps(value)}]

    # Default fallback
    return [{"type": "text", "text": str(value)}]


def _map_type_to_json_schema_type(val_type: str) -> str:
    """
    Map Arcade value types to JSON schema types.

    Args:
        val_type: The Arcade value type as a string.

    Returns:
        The corresponding JSON schema type as a string.
    """
    mapping: dict[str, str] = {
        "string": "string",
        "integer": "integer",
        "number": "number",
        "boolean": "boolean",
        "json": "object",
        "array": "array",
    }
    return mapping.get(val_type, "string")



from arcade_evals import (
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    NumericCritic,
    SimilarityCritic,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_search
from arcade_search.tools import search_google

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.8,
    warn_threshold=0.9,
)

catalog = ToolCatalog()
# Register the Google Search tool
catalog.add_module(arcade_search)


@tool_eval()
def google_search_eval_suite() -> EvalSuite:
    """Create an evaluation suite for the Google Search tool."""
    suite = EvalSuite(
        name="Google Search Tool Evaluation",
        system_message="You are an AI assistant that can perform web searches using the provided tools.",
        catalog=catalog,
        rubric=rubric,
    )

    # Simple search query with default results
    suite.add_case(
        name="Simple search query with default results",
        user_message="Search for 'Climate change effects on polar bears' on Google.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_google,
                args={
                    "query": "Climate change effects on polar bears",
                    "n_results": 5,
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="query", weight=1.0),
        ],
    )

    # Search query with specific number of results
    suite.add_case(
        name="Search query with specific number of results",
        user_message="Find the top 3 articles about quantum computing.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_google,
                args={
                    "query": "articles about quantum computing",
                    "n_results": 3,
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="query", weight=0.7),
            NumericCritic(
                critic_field="n_results",
                weight=0.3,
                value_range=(1, 100),
            ),
        ],
    )

    # Search query with 'n' results specified in words
    suite.add_case(
        name="Search query with 'n' results specified in words",
        user_message="Give me five recipes for vegan lasagna.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_google,
                args={
                    "query": "recipes for vegan lasagna",
                    "n_results": 5,
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="query", weight=0.7),
            NumericCritic(
                critic_field="n_results",
                weight=0.3,
                value_range=(1, 100),
            ),
        ],
    )

    # Ambiguous number of results
    suite.add_case(
        name="Ambiguous number of results",
        user_message="Find articles about climate change impacts 10.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_google,
                args={
                    "query": "articles about climate change impacts 10",
                    "n_results": 5,
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="query", weight=1.0),
        ],
    )

    # Search query with multiple instructions
    suite.add_case(
        name="Search query with multiple instructions",
        user_message="Search for the latest news on electric cars, and tell me about Tesla's new model.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_google,
                args={
                    "query": "latest news on electric cars",
                    "n_results": 5,
                },
            ),
            ExpectedToolCall(
                func=search_google,
                args={
                    "query": "Tesla's new model",
                    "n_results": 5,
                },
            ),
        ],
        critics=[
            SimilarityCritic(critic_field="query", weight=1.0),
        ],
    )

    # Search with stop words and filler words
    suite.add_case(
        name="Search with stop words and filler words",
        user_message="Could you please search for the best ways to learn French?",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_google,
                args={
                    "query": "best ways to learn French",
                    "n_results": 5,
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="query", weight=1.0),
        ],
    )

    # No clear query given
    suite.add_case(
        name="No clear query given",
        user_message="Find it for me.",
        expected_tool_calls=[],
        critics=[],
    )

    # Search query with special characters
    suite.add_case(
        name="Search query with special characters",
        user_message="Find me '@OpenAI's latest research papers'",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_google,
                args={
                    "query": "@OpenAI's latest research papers",
                    "n_results": 5,
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="query", weight=1.0),
        ],
    )

    # Search query with complex instructions
    suite.add_case(
        name="Search query with complex instructions",
        user_message="I need information about the impact of deforestation in the Amazon over the past decade.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_google,
                args={
                    "query": "impact of deforestation in the Amazon over the past decade",
                    "n_results": 5,
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="query", weight=1.0),
        ],
    )

    # Search query in a different language
    suite.add_case(
        name="Search query in a different language",
        user_message="Busca información sobre la economía de España.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_google,
                args={
                    "query": "economía de España",
                    "n_results": 5,
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="query", weight=1.0),
        ],
    )

    # Search query with numeric data
    suite.add_case(
        name="Search query with numeric data",
        user_message="What was the population of Japan in 2020?",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_google,
                args={
                    "query": "population of Japan in 2020",
                    "n_results": 5,
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="query", weight=1.0),
        ],
    )

    return suite



import asyncio
import logging
import os
import sys
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from functools import partial
from importlib.metadata import version as get_pkg_version
from pathlib import Path
from typing import Any

import fastapi
import uvicorn

# Watchfiles is used under the hood by Uvicorn's reload feature.
# Importing watchfiles here is an explicit acknowledgement that it needs to be installed
import watchfiles  # noqa: F401
from arcade_core.telemetry import OTELHandler
from arcade_core.toolkit import Toolkit, get_package_directory
from arcade_serve.fastapi.worker import FastAPIWorker
from loguru import logger
from rich.console import Console

from arcade_cli.constants import ARCADE_CONFIG_PATH
from arcade_cli.utils import (
    build_tool_catalog,
    discover_toolkits,
    load_dotenv,
)

console = Console(width=70, color_system="auto")


# App factory for Uvicorn reload
def create_arcade_app() -> fastapi.FastAPI:
    # TODO: Find a better way to pass these configs to factory used for reload
    debug_mode = os.environ.get("ARCADE_WORKER_SECRET", "dev") == "dev"
    otel_enabled = os.environ.get("ARCADE_OTEL_ENABLE", "False").lower() == "true"
    auth_for_reload = not debug_mode

    # Call setup_logging here to ensure Uvicorn worker processes also get Loguru formatting
    # for all standard library loggers.
    # The log_level for Uvicorn itself is set via uvicorn.run(log_level=...),
    # this call primarily aims to capture third-party library logs into Loguru.
    setup_logging(log_level=logging.DEBUG if debug_mode else logging.INFO, mcp_mode=False)

    logger.info(f"Debug: {debug_mode}, OTEL: {otel_enabled}, Auth Disabled: {auth_for_reload}")
    version = get_pkg_version("arcade-ai")
    toolkits = discover_toolkits()

    logger.info("Registered toolkits:")
    for toolkit in toolkits:
        logger.info(
            f"  - {toolkit.name}: {sum(len(tools) for tools in toolkit.tools.values())} tools"
        )

    otel_handler = OTELHandler(
        enable=otel_enabled,
        log_level=logging.DEBUG if debug_mode else logging.INFO,
    )

    custom_lifespan = partial(lifespan, otel_handler=otel_handler, enable_otel=otel_enabled)

    app = fastapi.FastAPI(
        title="Arcade Worker",
        description="A worker for the Arcade platform.",
        version=version,
        docs_url="/docs" if debug_mode else None,
        redoc_url="/redoc" if debug_mode else None,
        openapi_url="/openapi.json" if debug_mode else None,
        lifespan=custom_lifespan,
    )
    otel_handler.instrument_app(app)

    secret = os.getenv("ARCADE_WORKER_SECRET", "dev")
    if secret == "dev" and not os.environ.get("ARCADE_WORKER_SECRET"):  # noqa: S105
        logger.warning("Using default 'dev' for ARCADE_WORKER_SECRET. Set this in production.")

    worker = FastAPIWorker(
        app=app,
        secret=secret,
        disable_auth=not debug_mode,  # TODO (Sam): possible unexpected behavior on reload here?
        otel_meter=otel_handler.get_meter(),
    )
    for tk in toolkits:
        worker.register_toolkit(tk)

    return app


def _run_mcp_stdio(
    toolkits: list[Toolkit], *, logging_enabled: bool, env_file: str | None = None
) -> None:
    """Launch an MCP stdio server; blocks until it exits."""

    from arcade_serve.mcp.stdio import StdioServer

    # Load env vars before launching server (explicit path, config path, cwd)
    if env_file:
        load_dotenv(env_file, override=False)
    else:
        for candidate in [Path(ARCADE_CONFIG_PATH) / "arcade.env", Path.cwd() / "arcade.env"]:
            if candidate.is_file():
                load_dotenv(candidate, override=False)
                break

    # Set up middleware configuration for stdio mode
    middleware_config = {
        "stdio_mode": True,  # Ensure logs go to stderr
    }

    catalog = build_tool_catalog(toolkits)
    server = StdioServer(
        catalog,
        enable_logging=logging_enabled,
        middleware_config=middleware_config,
    )

    try:
        asyncio.run(server.run())
    except KeyboardInterrupt:
        logger.info("MCP server stopped by user.")
    except Exception as exc:
        logger.exception("Error while running MCP server: %s", exc)
        raise
    finally:
        logger.info("Shutting down Server")
        logger.complete()
        logger.remove()


def _run_fastapi_server(
    host: str,
    port: int,
    workers_param: int,
    timeout_keep_alive: int,
    reload: bool,
    toolkits_for_reload_dirs: list[Toolkit] | None,
    debug_flag: bool,
) -> None:
    app_import_string = "arcade_cli.serve:create_arcade_app"
    reload_dirs_str_list: list[str] | None = None

    if reload:
        current_reload_dirs_paths = []
        if toolkits_for_reload_dirs:
            for tk in toolkits_for_reload_dirs:
                try:
                    package_dir_str = get_package_directory(tk.package_name)
                    current_reload_dirs_paths.append(Path(package_dir_str))
                except Exception as e:
                    logger.warning(f"Error getting reload path for toolkit {tk.name}: {e}")

        serve_py_dir_path = Path(__file__).resolve().parent
        current_reload_dirs_paths.append(serve_py_dir_path)

        if current_reload_dirs_paths:
            reload_dirs_str_list = [str(p) for p in current_reload_dirs_paths]
            logger.debug(f"Uvicorn reload_dirs: {reload_dirs_str_list}")

    effective_workers = 1 if reload else workers_param
    log_level_str = logging.getLevelName(logging.DEBUG if debug_flag else logging.INFO).lower()

    logger.debug(
        f"Calling uvicorn.run with app='{app_import_string}', factory=True, host='{host}', port={port}, "
        f"workers={effective_workers}, reload={reload}, log_level='{log_level_str}'"
    )

    uvicorn.run(
        app_import_string,
        factory=True,
        host=host,
        port=port,
        workers=effective_workers,
        log_config=None,
        log_level=log_level_str,
        reload=reload,
        reload_dirs=reload_dirs_str_list,
        lifespan="on",
        timeout_keep_alive=timeout_keep_alive,
    )


class RichInterceptHandler(logging.Handler):
    def emit(self, record: logging.LogRecord) -> None:
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = str(record.levelno)
        logger.opt(exception=record.exc_info).log(level, record.getMessage())


def setup_logging(log_level: int = logging.INFO, mcp_mode: bool = False) -> None:
    """Loguru and intercepts standard logging."""
    # Set our handler on root
    logging.root.handlers = [RichInterceptHandler()]
    logging.root.setLevel(log_level)

    # For all existing loggers, remove their handlers and make them propagate to root.
    for name in list(logging.root.manager.loggerDict.keys()):
        existing_logger = logging.getLogger(name)
        existing_logger.handlers = []
        existing_logger.propagate = True

    # clear existing loguru handlers to keep worker logging behavior clean
    # and consistent despite toolkit logging changes
    logger.remove()

    # set sink destination based on mode
    # MCP stdio needs to write to stderr to avoid interfering with capture
    sink_destination = sys.stderr if mcp_mode else sys.stdout

    if log_level == logging.DEBUG:
        format_string = "<level>{level}</level> | <green>{time:HH:mm:ss}</green> | <cyan>{name}:{file}:{line: <4}</cyan> | <level>{message}</level>"
    else:
        format_string = (
            "<level>{level}</level> | <green>{time:HH:mm:ss}</green> | <level>{message}</level>"
        )

    logger.configure(
        handlers=[
            {
                "sink": sink_destination,
                "colorize": True,
                "level": log_level,
                "format": format_string,
                "enqueue": True,  # non-blocking logging
                "diagnose": False,  # disable detailed logging TODO: make this configurable
            }
        ]
    )


@asynccontextmanager
async def lifespan(
    app: fastapi.FastAPI, otel_handler: OTELHandler | None = None, enable_otel: bool = False
) -> AsyncGenerator[None, None]:
    try:
        logger.debug(f"Server lifespan startup. OTEL enabled: {enable_otel}")
        yield
    except (asyncio.CancelledError, KeyboardInterrupt):
        logger.debug("Server lifespan cancelled.")
        raise
    finally:
        logger.debug(f"Server lifespan shutdown. OTEL enabled: {enable_otel}")
        if enable_otel and otel_handler:
            otel_handler.shutdown()
        await logger.complete()
        logger.remove()
        logger.debug("Server lifespan shutdown complete.")


def serve_default_worker(
    host: str = "127.0.0.1",
    port: int = 8002,
    disable_auth: bool = False,
    workers: int = 1,
    timeout_keep_alive: int = 5,
    enable_otel: bool = False,
    debug: bool = False,
    mcp: bool = False,
    reload: bool = False,
    **kwargs: Any,
) -> None:
    # Initial logging setup for the main `arcade serve` process itself.
    # The Uvicorn worker processes will call setup_logging() again via create_arcade_app().
    setup_logging(log_level=logging.DEBUG if debug else logging.INFO, mcp_mode=mcp)

    if mcp:
        logger.info("MCP mode selected.")
        toolkits_for_mcp = discover_toolkits()
        _run_mcp_stdio(
            toolkits_for_mcp, logging_enabled=not debug, env_file=kwargs.pop("env_file", None)
        )
        return

    logger.info("FastAPI mode selected. Configuring for Uvicorn with app factory.")
    os.environ["ARCADE_DEBUG_MODE"] = str(debug)
    os.environ["ARCADE_OTEL_ENABLE"] = str(enable_otel)
    os.environ["ARCADE_DISABLE_AUTH"] = str(disable_auth)

    toolkits_for_reload_dirs: list[Toolkit] | None = None
    if reload:
        # This discovery is only to tell the main Uvicorn reloader process which project dirs to watch.
        # The actual app running in the worker will do its own discovery via create_arcade_app.
        toolkits_for_reload_dirs = discover_toolkits()
        logger.debug(
            f"Reload mode: Uvicorn to watch {len(toolkits_for_reload_dirs) if toolkits_for_reload_dirs else 0} directories."
        )

    _run_fastapi_server(
        host=host,
        port=port,
        workers_param=workers,
        timeout_keep_alive=timeout_keep_alive,
        reload=reload,
        toolkits_for_reload_dirs=toolkits_for_reload_dirs,
        debug_flag=debug,
    )
    logger.info("Arcade serve process finished.")



from arcade_math.tools.arithmetic import (
    add,
    divide,
    mod,
    multiply,
    subtract,
    sum_list,
    sum_range,
)
from arcade_math.tools.exponents import (
    log,
    power,
)
from arcade_math.tools.miscellaneous import (
    abs_val,
    factorial,
    sqrt,
)
from arcade_math.tools.random import (
    generate_random_float,
    generate_random_int,
)
from arcade_math.tools.rational import (
    gcd,
    lcm,
)
from arcade_math.tools.rounding import (
    ceil,
    floor,
    round_num,
)
from arcade_math.tools.statistics import (
    avg,
    median,
)
from arcade_math.tools.trigonometry import (
    deg_to_rad,
    rad_to_deg,
)

__all__ = [
    "add",
    "subtract",
    "multiply",
    "divide",
    "sum_list",
    "sum_range",
    "mod",
    "log",
    "power",
    "abs_val",
    "factorial",
    "sqrt",
    "generate_random_float",
    "generate_random_int",
    "gcd",
    "lcm",
    "ceil",
    "floor",
    "round_num",
    "avg",
    "median",
    "deg_to_rad",
    "rad_to_deg",
]



from datetime import timedelta

import pytest
import pytz
from arcade_evals import (
    BinaryCritic,
    DatetimeCritic,
    NoneCritic,
    NumericCritic,
    SimilarityCritic,
)
from arcade_evals.errors import WeightError
from dateutil import parser


# Test NoneCritic initialization
@pytest.mark.parametrize("weight, expected_weight", [(0.0, 0.0), (0.5, 0.0)])
def test_none_critic_initialization(weight, expected_weight):
    field_name = "my_field"

    critic = NoneCritic(weight=weight, critic_field=field_name)
    assert critic.weight == expected_weight
    assert critic.critic_field == field_name


# Test NoneCritic.evaluate()
def test_none_critic_evaluate():
    critic = NoneCritic(critic_field="my_field")
    result = critic.evaluate(expected="expected_value", actual="actual_value")
    assert result["match"] is None
    assert result["score"] == 0.0
    assert result["is_criticized"] is False


# Test BinaryCritic.evaluate()
@pytest.mark.parametrize(
    "expected, actual, weight, expected_match, expected_score",
    [
        ("value", "value", 1.0, True, 1.0),
        ("value", "different", 1.0, False, 0.0),
        (10, 10, 0.5, True, 0.5),
        (10, 20, 0.5, False, 0.0),
    ],
)
def test_binary_critic_evaluate(expected, actual, weight, expected_match, expected_score):
    """
    Test the BinaryCritic's evaluate method to ensure it correctly computes
    the match and score based on expected and actual values.
    """
    critic = BinaryCritic(critic_field="test_field", weight=weight)
    result = critic.evaluate(expected=expected, actual=actual)
    assert result["match"] == expected_match
    assert result["score"] == expected_score


# Test NumericCritic.evaluate()
@pytest.mark.parametrize(
    "expected, actual, value_range, weight, match_threshold, expected_match, expected_score",
    [
        (5, 5, (0, 10), 1.0, 0.8, True, 1.0),
        (5, 6, (0, 10), 1.0, 0.8, True, 0.9),
        (0, 10, (0, 10), 1.0, 0.8, False, 0.0),
        (2, 8, (0, 10), 1.0, 0.5, False, 0.4),
        (50, 60, (0, 100), 0.5, 0.9, True, 0.45),
    ],
)
def test_numeric_critic_evaluate(
    expected, actual, value_range, weight, match_threshold, expected_match, expected_score
):
    """
    Test the NumericCritic's evaluate method to ensure it calculates
    the correct score based on the proportion of the difference between
    expected and actual values within a specified range.
    """
    critic = NumericCritic(
        critic_field="number",
        weight=weight,
        value_range=value_range,
        match_threshold=match_threshold,
    )
    result = critic.evaluate(expected=expected, actual=actual)
    assert result["match"] == expected_match
    assert pytest.approx(result["score"], 0.01) == expected_score


# Test SimilarityCritic.evaluate()
@pytest.mark.parametrize(
    "expected, actual, weight, similarity_threshold, expected_match, min_expected_score",
    [
        ("hello world", "hello world", 1.0, 0.8, True, 1.0),
        ("hello world", "hello", 1.0, 0.8, False, 0.0),
        ("The quick brown fox", "The quick brown fox jumps over the lazy dog", 1.0, 0.5, True, 0.5),
        ("data science", "machine learning", 0.5, 0.3, False, 0.0),
    ],
)
def test_similarity_critic_evaluate(
    expected, actual, weight, similarity_threshold, expected_match, min_expected_score
):
    """
    Test the SimilarityCritic's evaluate method to ensure it computes
    the similarity score between expected and actual strings and determines
    the match correctly based on the similarity threshold.
    """
    critic = SimilarityCritic(
        critic_field="text",
        weight=weight,
        similarity_threshold=similarity_threshold,
    )
    result = critic.evaluate(expected=expected, actual=actual)
    assert result["match"] == expected_match
    assert result["score"] >= min_expected_score
    assert result["score"] >= 0.0
    assert result["score"] <= weight + 1e-6  # Allow a small epsilon for floating-point comparison


# Test that WeightError is raised for invalid critic weights
@pytest.mark.parametrize(
    "critic_class, weight",
    [
        (BinaryCritic, -0.1),
        (BinaryCritic, 1.1),
        (NumericCritic, -0.5),
        (SimilarityCritic, 1.5),
    ],
)
def test_critic_invalid_weight(critic_class, weight):
    """
    Test that initializing a critic with an invalid weight raises a WeightError.
    """
    with pytest.raises(WeightError):
        if critic_class == NumericCritic:
            critic_class(critic_field="test_field", weight=weight, value_range=(0, 1))
        elif critic_class == SimilarityCritic:
            critic_class(critic_field="test_field", weight=weight)
        else:
            critic_class(critic_field="test_field", weight=weight)


# Test NumericCritic with invalid value range
def test_numeric_critic_invalid_range():
    """
    Test that initializing a NumericCritic with an invalid value range raises a ValueError.
    """
    with pytest.raises(ValueError):
        NumericCritic(critic_field="number", weight=1.0, value_range=(10, 0))  # Invalid range


# Test SimilarityCritic with unsupported metric
def test_similarity_critic_unsupported_metric():
    """
    Test that initializing a SimilarityCritic with an unsupported metric raises a ValueError.
    """
    with pytest.raises(ValueError):
        SimilarityCritic(critic_field="text", weight=1.0, metric="unsupported_metric")


# Test DatetimeCritic
# Parameterized tests for DatetimeCritic with various datetime formats and default timezones
@pytest.mark.parametrize(
    "critic_params, expected, actual, expected_match, expected_score",
    [
        # Test with time component and timezone
        (
            {"critic_field": "start_datetime", "weight": 1.0},
            "2024-09-26T12:00:00-07:00",
            "2024-09-26T12:00:00-07:00",
            True,
            1.0,
        ),
        # Test without time component (dates only)
        (
            {"critic_field": "start_datetime", "weight": 1.0},
            "2024-09-26",
            "2024-09-26",
            True,
            1.0,
        ),
        # Test with and without timezone (assumes UTC)
        (
            {"critic_field": "start_datetime", "weight": 1.0},
            "2024-09-26T12:00:00Z",
            "2024-09-26T12:00:00",
            True,
            1.0,
        ),
        # Test naive datetimes
        (
            {"critic_field": "start_datetime", "weight": 1.0},
            "2024-09-26T12:00:00",
            "2024-09-26T12:00:00",
            True,
            1.0,
        ),
    ],
)
def test_datetime_critic_basic(critic_params, expected, actual, expected_match, expected_score):
    """
    Test DatetimeCritic with various datetime formats and default timezones.
    """
    critic = DatetimeCritic(**critic_params)
    result = critic.evaluate(expected, actual)
    assert result["match"] == expected_match
    assert result["score"] == expected_score


# Parameterized tests for DatetimeCritic's handling of tolerances and max differences
@pytest.mark.parametrize(
    "critic_params, expected, actual, expected_match, expected_score_func",
    [
        # Test time difference within tolerance
        (
            {"critic_field": "start_datetime", "weight": 1.0, "tolerance": timedelta(seconds=60)},
            "2024-09-26T12:00:00",
            "2024-09-26T12:00:30",
            True,
            lambda critic: critic.weight,
        ),
        # Test time difference outside tolerance but within max_difference
        (
            {
                "critic_field": "start_datetime",
                "weight": 1.0,
                "tolerance": timedelta(seconds=60),
                "max_difference": timedelta(minutes=5),
            },
            "2024-09-26T12:00:00",
            "2024-09-26T12:04:00",
            False,
            lambda critic: critic.weight * (1 - (240 / 300)),
        ),
        # Test time difference exceeds max_difference
        (
            {
                "critic_field": "start_datetime",
                "weight": 1.0,
                "max_difference": timedelta(minutes=5),
            },
            "2024-09-26T12:00:00",
            "2024-09-26T12:10:00",
            False,
            lambda critic: 0.0,
        ),
    ],
)
def test_datetime_critic_tolerances(
    critic_params, expected, actual, expected_match, expected_score_func
):
    """
    Test DatetimeCritic's handling of tolerances and max differences.
    """
    critic = DatetimeCritic(**critic_params)
    result = critic.evaluate(expected, actual)
    assert result["match"] == expected_match
    expected_score = expected_score_func(critic)
    assert pytest.approx(result["score"], abs=1e-6) == expected_score


def test_datetime_critic_naive_and_timezone_aware():
    """
    Test DatetimeCritic when comparing naive and timezone-aware datetimes.
    """
    critic = DatetimeCritic(critic_field="start_datetime", weight=1.0)
    expected = "2024-09-26T12:00:00Z"
    actual = "2024-09-26T07:00:00"
    result = critic.evaluate(expected, actual)
    assert result["match"] is False

    # Compute expected score based on time difference
    expected_dt = parser.parse(expected)
    actual_dt = parser.parse(actual)
    if actual_dt.tzinfo is None:
        actual_dt = pytz.utc.localize(actual_dt)
    if expected_dt.tzinfo is None:
        expected_dt = pytz.utc.localize(expected_dt)

    time_diff_seconds = abs((expected_dt - actual_dt).total_seconds())
    if time_diff_seconds <= critic.tolerance.total_seconds():
        expected_score = critic.weight
    elif time_diff_seconds >= critic.max_difference.total_seconds():
        expected_score = 0.0
    else:
        ratio = 1 - (time_diff_seconds / critic.max_difference.total_seconds())
        expected_score = critic.weight * ratio

    assert pytest.approx(result["score"], abs=1e-6) == expected_score



import base64
import json
from typing import Annotated

from arcade_tdk import ToolContext, ToolMetadataKey, tool
from arcade_tdk.auth import Google
from arcade_tdk.errors import ToolExecutionError
from googleapiclient.errors import HttpError

from arcade_google_drive.enums import OrderBy
from arcade_google_drive.templates import optional_file_picker_instructions_template
from arcade_google_drive.utils import (
    build_drive_service,
    build_file_tree,
    build_file_tree_request_params,
)


@tool(
    requires_auth=Google(
        scopes=["https://www.googleapis.com/auth/drive.file"],
    ),
    requires_metadata=[ToolMetadataKey.CLIENT_ID, ToolMetadataKey.COORDINATOR_URL],
)
async def get_file_tree_structure(
    context: ToolContext,
    include_shared_drives: Annotated[
        bool, "Whether to include shared drives in the file tree structure. Defaults to False."
    ] = False,
    restrict_to_shared_drive_id: Annotated[
        str | None,
        "If provided, only include files from this shared drive in the file tree structure. "
        "Defaults to None, which will include files and folders from all drives.",
    ] = None,
    include_organization_domain_documents: Annotated[
        bool,
        "Whether to include documents from the organization's domain. This is applicable to admin "
        "users who have permissions to view organization-wide documents in a Google Workspace "
        "account. Defaults to False.",
    ] = False,
    order_by: Annotated[
        list[OrderBy] | None,
        "Sort order. Defaults to listing the most recently modified documents first",
    ] = None,
    limit: Annotated[
        int | None,
        "The number of files and folders to list. Defaults to None, "
        "which will list all files and folders.",
    ] = None,
) -> Annotated[
    dict,
    "A dictionary containing the file/folder tree structure in the user's Google Drive",
]:
    """
    Get the file/folder tree structure of the user's Google Drive.
    """
    service = build_drive_service(context.get_auth_token_or_empty())

    keep_paginating = True
    page_token = None
    files = {}
    file_tree: dict[str, list[dict]] = {"My Drive": []}

    params = build_file_tree_request_params(
        order_by,
        page_token,
        limit,
        include_shared_drives,
        restrict_to_shared_drive_id,
        include_organization_domain_documents,
    )

    while keep_paginating:
        # Get a list of files
        results = service.files().list(**params).execute()

        # Update page token
        page_token = results.get("nextPageToken")
        params["pageToken"] = page_token
        keep_paginating = page_token is not None

        for file in results.get("files", []):
            files[file["id"]] = file

    if not files:
        return {"drives": []}

    file_tree = build_file_tree(files)

    drives = []

    for drive_id, files in file_tree.items():  # type: ignore[assignment]
        if drive_id == "My Drive":
            drive = {"name": "My Drive", "children": files}
        else:
            try:
                drive_details = service.drives().get(driveId=drive_id).execute()
                drive_name = drive_details.get("name", "Shared Drive (name unavailable)")
            except HttpError as e:
                drive_name = (
                    f"Shared Drive (name unavailable: 'HttpError {e.status_code}: {e.reason}')"
                )

            drive = {"name": drive_name, "id": drive_id, "children": files}

        drives.append(drive)

    file_picker_response = generate_google_file_picker_url(
        context,
    )

    return {
        "drives": drives,
        "file_picker": {
            "url": file_picker_response["url"],
            "llm_instructions": optional_file_picker_instructions_template.format(
                url=file_picker_response["url"]
            ),
        },
    }


@tool(
    requires_auth=Google(),
    requires_metadata=[ToolMetadataKey.CLIENT_ID, ToolMetadataKey.COORDINATOR_URL],
)
def generate_google_file_picker_url(
    context: ToolContext,
) -> Annotated[dict, "Google File Picker URL for user file selection and permission granting"]:
    """Generate a Google File Picker URL for user-driven file selection and authorization.

    This tool generates a URL that directs the end-user to a Google File Picker interface where
    where they can select or upload Google Drive files. Users can grant permission to access their
    Drive files, providing a secure and authorized way to interact with their files.

    This is particularly useful when prior tools (e.g., those accessing or modifying
    Google Docs, Google Sheets, etc.) encountered failures due to file non-existence
    (Requested entity was not found) or permission errors. Once the user completes the file
    picker flow, the prior tool can be retried.
    """
    client_id = context.get_metadata(ToolMetadataKey.CLIENT_ID)
    client_id_parts = client_id.split("-")
    if not client_id_parts:
        raise ToolExecutionError(
            message="Invalid Google Client ID",
            developer_message=f"Google Client ID '{client_id}' is not valid",
        )
    app_id = client_id_parts[0]
    cloud_coordinator_url = context.get_metadata(ToolMetadataKey.COORDINATOR_URL).strip("/")

    config = {
        "auth": {
            "client_id": client_id,
            "app_id": app_id,
        },
    }
    config_json = json.dumps(config)
    config_base64 = base64.urlsafe_b64encode(config_json.encode("utf-8")).decode("utf-8")
    url = f"{cloud_coordinator_url}/google/drive_picker?config={config_base64}"

    return {
        "url": url,
        "llm_instructions": (
            "Instruct the user to click the following link to open the Google Drive File Picker. "
            "This will allow them to select files and grant access permissions: {url}"
        ),
    }



import sys
import types
from typing import Annotated, Any

import pytest
from arcade_core.catalog import ToolCatalog
from arcade_serve.mcp import server as mcp_server
from arcade_serve.mcp.types import (
    CallToolRequest,
    CancelRequest,
    InitializeRequest,
    ListToolsRequest,
    PingRequest,
)
from arcade_tdk import tool

# ---------------------------------------------------------------------------
# Test helpers / stubs
# ---------------------------------------------------------------------------


class _FakeAuth:
    async def authorize(self, auth_requirement: Any, user_id: str):
        """Return an object that mimics AuthorizationResponse with completed status."""

        class _Ctx:  # minimal stub
            token = "dummy-token"  # noqa: S105

        class _Resp:  # pylint: disable=too-few-public-methods
            status = "completed"
            url = ""
            context = _Ctx()

        return _Resp()


class _FakeArcade:  # pylint: disable=too-few-public-methods
    def __init__(self, **_: Any):
        self.auth = _FakeAuth()


# Ensure that the AsyncArcade & ArcadeError symbols inside server.py point to our stubs.
pytestmark = pytest.mark.asyncio


@pytest.fixture(autouse=True)
def _patch_arcadepy(monkeypatch):
    """Patch the external `arcadepy` dependency used by mcp.server."""

    # Patch the imported symbols on the already-imported server module
    monkeypatch.setattr(mcp_server, "AsyncArcade", _FakeArcade, raising=True)
    monkeypatch.setattr(mcp_server, "ArcadeError", Exception, raising=True)

    # Provide a dummy `arcadepy` module in sys.modules for any other importers
    fake_arcadepy = types.ModuleType("arcadepy")
    fake_arcadepy.AsyncArcade = _FakeArcade  # type: ignore[attr-defined]
    fake_arcadepy.ArcadeError = Exception  # type: ignore[attr-defined]
    sys.modules["arcadepy"] = fake_arcadepy

    yield

    # Cleanup
    sys.modules.pop("arcadepy", None)


# ---------------------------------------------------------------------------
# Fixtures for a sample tool / catalog / server
# ---------------------------------------------------------------------------


@tool
def multiply(a: Annotated[int, "a"], b: Annotated[int, "b"]) -> Annotated[int, "result"]:
    """Return the product of *a* and *b*."""

    return a * b


@pytest.fixture(scope="module")
def sample_catalog():
    catalog = ToolCatalog()
    catalog.add_tool(multiply, "test_toolkit")
    return catalog


@pytest.fixture()
def server(sample_catalog):
    # MCPServer constructor is synchronous, so fixture need not be async
    return mcp_server.MCPServer(sample_catalog, enable_logging=False)


# ---------------------------------------------------------------------------
# Tests
# ---------------------------------------------------------------------------


async def test_handle_ping(server):
    req = PingRequest(id=123)
    resp = await server._handle_ping(req)  # pylint: disable=protected-access
    assert resp.id == 123
    assert resp.result == {"pong": True}


async def test_handle_initialize(server):
    req = InitializeRequest(id=1)
    resp = await server._handle_initialize(req)  # pylint: disable=protected-access
    assert resp.id == 1
    assert resp.result.protocolVersion == mcp_server.MCP_PROTOCOL_VERSION
    assert resp.result.serverInfo.name.startswith("Arcade")


async def test_handle_list_tools(server):
    req = ListToolsRequest(id=99)
    resp = await server._handle_list_tools(req)  # pylint: disable=protected-access
    assert resp.id == 99
    # Should list our sample tool only
    tool_names = [t.name for t in resp.result.tools]
    assert "TestToolkit_Multiply" in tool_names  # toolkit + "_" + tool


async def test_handle_call_tool_success(server):
    req = CallToolRequest(
        id="call-1",
        params={
            "name": "TestToolkit_Multiply",
            "input": {"a": 6, "b": 7},
        },
    )
    resp = await server._handle_call_tool(req, user_id="tester@example.com")  # pylint: disable=protected-access

    assert resp.id == "call-1"
    # convert_to_mcp_content wraps primitives in list-of-dicts
    assert resp.result.content == [{"type": "text", "text": "42"}]


async def test_send_response_dict(server, monkeypatch):
    """_send_response should JSON-serialize plain dictionaries."""

    sent: list[str] = []

    class _Write:
        async def send(self, msg):
            sent.append(msg)

    await server._send_response(_Write(), {"foo": "bar"})  # pylint: disable=protected-access

    assert sent and sent[0].strip() == '{"foo": "bar"}'


async def test_handle_cancel(server):
    req = CancelRequest(id=77, params={"id": "abc"})
    resp = await server._handle_cancel(req)  # pylint: disable=protected-access
    assert resp.result == {"ok": True}



import logging
from typing import Any

from google.oauth2.credentials import Credentials
from googleapiclient.discovery import Resource, build

from arcade_google_docs.enum import Corpora, OrderBy

## Set up basic configuration for logging to the console with DEBUG level and a specific format.
logging.basicConfig(
    level=logging.DEBUG,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)

logger = logging.getLogger(__name__)


def build_docs_service(auth_token: str | None) -> Resource:  # type: ignore[no-any-unimported]
    """
    Build a Drive service object.
    """
    auth_token = auth_token or ""
    return build("docs", "v1", credentials=Credentials(auth_token))


def build_drive_service(auth_token: str | None) -> Resource:  # type: ignore[no-any-unimported]
    """
    Build a Drive service object.
    """
    auth_token = auth_token or ""
    return build("drive", "v3", credentials=Credentials(auth_token))


def build_files_list_params(
    mime_type: str,
    page_size: int,
    order_by: list[OrderBy],
    pagination_token: str | None,
    include_shared_drives: bool,
    search_only_in_shared_drive_id: str | None,
    include_organization_domain_documents: bool,
    document_contains: list[str] | None = None,
    document_not_contains: list[str] | None = None,
) -> dict[str, Any]:
    query = build_files_list_query(
        mime_type=mime_type,
        document_contains=document_contains,
        document_not_contains=document_not_contains,
    )

    params = {
        "q": query,
        "pageSize": page_size,
        "orderBy": ",".join([item.value for item in order_by]),
        "pageToken": pagination_token,
    }

    if (
        include_shared_drives
        or search_only_in_shared_drive_id
        or include_organization_domain_documents
    ):
        params["includeItemsFromAllDrives"] = "true"
        params["supportsAllDrives"] = "true"

    if search_only_in_shared_drive_id:
        params["driveId"] = search_only_in_shared_drive_id
        params["corpora"] = Corpora.DRIVE.value

    if include_organization_domain_documents:
        params["corpora"] = Corpora.DOMAIN.value

    params = remove_none_values(params)

    return params


def build_files_list_query(
    mime_type: str,
    document_contains: list[str] | None = None,
    document_not_contains: list[str] | None = None,
) -> str:
    query = [f"(mimeType = '{mime_type}' and trashed = false)"]

    if isinstance(document_contains, str):
        document_contains = [document_contains]

    if isinstance(document_not_contains, str):
        document_not_contains = [document_not_contains]

    if document_contains:
        for keyword in document_contains:
            name_contains = keyword.replace("'", "\\'")
            full_text_contains = keyword.replace("'", "\\'")
            keyword_query = (
                f"(name contains '{name_contains}' or fullText contains '{full_text_contains}')"
            )
            query.append(keyword_query)

    if document_not_contains:
        for keyword in document_not_contains:
            name_not_contains = keyword.replace("'", "\\'")
            full_text_not_contains = keyword.replace("'", "\\'")
            keyword_query = (
                f"(name not contains '{name_not_contains}' and "
                f"fullText not contains '{full_text_not_contains}')"
            )
            query.append(keyword_query)

    return " and ".join(query)


def remove_none_values(params: dict) -> dict:
    """
    Remove None values from a dictionary.
    :param params: The dictionary to clean
    :return: A new dictionary with None values removed
    """
    return {k: v for k, v in params.items() if v is not None}



from arcade_core.errors import RetryableToolError, ToolExecutionError, ToolRuntimeError

__all__ = [
    "RetryableToolError",
    "SDKError",
    "ToolExecutionError",
    "ToolRuntimeError",
    "WeightError",
]


class SDKError(Exception):
    """Base class for all SDK errors."""


class WeightError(SDKError):
    """Raised when the critic weights do not abide by SDK weight constraints."""



import asyncio

import pytest
from arcade_core.auth import AuthProviderType, Google
from arcade_tdk import tool
from arcade_tdk.auth import OAuth2


def test_sync_function():
    """
    Ensures a function will run when decorated by @tool
    """

    @tool
    def sync_func(x, y):
        return x + y

    result = sync_func(1, 2)
    assert result == 3


@pytest.mark.asyncio
async def test_async_function():
    """
    Ensures an async function will run when decorated by @tool
    """

    @tool
    async def async_func(x, y):
        await asyncio.sleep(0)
        return x + y

    result = await async_func(1, 2)
    assert result == 3


@pytest.mark.parametrize(
    "auth_class, auth_kwargs, expected_provider_id, expected_id",
    [
        (
            OAuth2,
            {"id": "my_example_provider123", "scopes": ["test_scope", "another.scope"]},
            None,
            "my_example_provider123",
        ),
        (Google, {"scopes": ["test_scope", "another.scope"]}, "google", None),
        (
            Google,
            {"id": "my_google_provider123", "scopes": ["test_scope", "another.scope"]},
            "google",
            "my_google_provider123",
        ),
    ],
)
def test_tool_decorator_with_auth_success(
    auth_class, auth_kwargs, expected_provider_id, expected_id
):
    @tool(
        name="TestTool",
        desc="Test description",
        requires_auth=auth_class(**auth_kwargs),
    )
    def test_tool(x, y):
        return x + y

    assert test_tool.__tool_name__ == "TestTool"
    assert test_tool.__tool_description__ == "Test description"
    assert test_tool.__tool_requires_auth__.provider_id == expected_provider_id
    assert test_tool.__tool_requires_auth__.provider_type == AuthProviderType.oauth2
    assert test_tool.__tool_requires_auth__.id == expected_id
    assert test_tool.__tool_requires_auth__.scopes == ["test_scope", "another.scope"]


@pytest.mark.parametrize(
    "auth_class, auth_kwargs",
    [
        (OAuth2, {"scopes": ["test_scope", "another.scope"]}),
        (
            OAuth2,
            {"provider_id": "my_example_provider123", "scopes": ["test_scope", "another.scope"]},
        ),
        (
            OAuth2,
            {
                "provider_id": "my_example_provider_id_123",
                "id": "my_example_id_123",
                "scopes": ["test_scope", "another.scope"],
            },
        ),
        (
            Google,
            {
                "provider_id": "my_example_provider_id_123",
                "scopes": ["test_scope", "another.scope"],
            },
        ),
        (
            Google,
            {
                "provider_id": "my_example_provider_id_123",
                "id": "my_example_id_123",
                "scopes": ["test_scope", "another.scope"],
            },
        ),
    ],
)
def test_tool_decorator_with_auth_failure(auth_class, auth_kwargs):
    with pytest.raises(TypeError):

        @tool(
            name="TestTool",
            desc="Test description",
            requires_auth=auth_class(**auth_kwargs),
        )
        def test_tool(x, y):
            return x + y


def test_tool_deprecated_ordering_no_auth():
    """
    Checks the behavior of @tool.deprecated when used before and after the @tool decorator.
    The order of the decorators should not matter.
    """
    message = "Deprecated: please use new_tool instead."

    @tool.deprecated(message)
    @tool
    def func_deprecated_after(x):
        """Test description for func_deprecated_after"""
        return x

    assert hasattr(func_deprecated_after, "__tool_deprecation_message__")
    assert func_deprecated_after.__tool_deprecation_message__ == message
    assert func_deprecated_after.__tool_name__ == "FuncDeprecatedAfter"
    assert (
        func_deprecated_after.__tool_description__ == "Test description for func_deprecated_after"
    )
    assert func_deprecated_after.__tool_requires_auth__ is None

    @tool
    @tool.deprecated(message)
    def func_deprecated_before(x):
        """Test description for func_deprecated_before"""
        return x

    assert hasattr(func_deprecated_before, "__tool_deprecation_message__")
    assert func_deprecated_before.__tool_deprecation_message__ == message
    assert func_deprecated_before.__tool_name__ == "FuncDeprecatedBefore"
    assert (
        func_deprecated_before.__tool_description__ == "Test description for func_deprecated_before"
    )
    assert func_deprecated_before.__tool_requires_auth__ is None


def test_tool_deprecated_ordering_with_auth():
    """
    Checks the behavior of @tool.deprecated when used with authentication.
    The order of the decorators should not matter.
    """
    message = "Deprecated: please use new_tool instead."

    @tool.deprecated(message)
    @tool(requires_auth=OAuth2(id="my_auth_id", scopes=["test_scope"]))
    def func_deprecated_after_auth(x):
        """Test description for func_deprecated_after_auth"""
        return x

    assert hasattr(func_deprecated_after_auth, "__tool_deprecation_message__")
    assert func_deprecated_after_auth.__tool_deprecation_message__ == message
    assert func_deprecated_after_auth.__tool_name__ == "FuncDeprecatedAfterAuth"
    assert (
        func_deprecated_after_auth.__tool_description__
        == "Test description for func_deprecated_after_auth"
    )
    assert func_deprecated_after_auth.__tool_requires_auth__ is not None

    @tool(requires_auth=OAuth2(id="my_auth_id", scopes=["test_scope"]))
    @tool.deprecated(message)
    def func_deprecated_before_auth(x):
        """Test description for func_deprecated_before_auth"""
        return x

    assert hasattr(func_deprecated_before_auth, "__tool_deprecation_message__")
    assert func_deprecated_before_auth.__tool_deprecation_message__ == message
    assert func_deprecated_before_auth.__tool_name__ == "FuncDeprecatedBeforeAuth"
    assert (
        func_deprecated_before_auth.__tool_description__
        == "Test description for func_deprecated_before_auth"
    )
    assert func_deprecated_before_auth.__tool_requires_auth__ is not None



from typing import Annotated, Any, cast

from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import Atlassian

import arcade_jira.cache as cache
from arcade_jira.client import JiraClient
from arcade_jira.exceptions import NotFoundError
from arcade_jira.utils import (
    add_pagination_to_response,
    clean_project_dict,
    remove_none_values,
)


@tool(requires_auth=Atlassian(scopes=["read:jira-work"]))
async def list_projects(
    context: ToolContext,
    limit: Annotated[
        int, "The maximum number of projects to return. Min of 1, Max of 50. Defaults to 50."
    ] = 50,
    offset: Annotated[
        int, "The number of projects to skip. Defaults to 0 (starts from the first project)"
    ] = 0,
) -> Annotated[dict[str, Any], "Information about the projects"]:
    """Browse projects available in Jira."""
    return cast(
        dict[str, Any], await search_projects(context, keywords=None, limit=limit, offset=offset)
    )


@tool(requires_auth=Atlassian(scopes=["read:jira-work"]))
async def search_projects(
    context: ToolContext,
    keywords: Annotated[
        str | None,
        "The keywords to search for projects. Matches against project name and key "
        "(case insensitive). Defaults to None (no keywords filter).",
    ] = None,
    limit: Annotated[
        int, "The maximum number of projects to return. Min of 1, Max of 50. Defaults to 50."
    ] = 50,
    offset: Annotated[
        int, "The number of projects to skip. Defaults to 0 (starts from the first project)"
    ] = 0,
) -> Annotated[dict[str, Any], "Information about the projects"]:
    """Get the details of all Jira projects."""
    limit = max(min(limit, 50), 1)
    client = JiraClient(context.get_auth_token_or_empty())
    api_response = await client.get(
        "/project/search",
        params=remove_none_values({
            "expand": ",".join([
                "description",
                "url",
            ]),
            "maxResults": limit,
            "startAt": offset,
            "query": keywords,
        }),
    )
    cloud_name = cache.get_cloud_name(context.get_auth_token_or_empty())
    projects = [clean_project_dict(project, cloud_name) for project in api_response["values"]]
    response = {
        "projects": projects,
        "isLast": api_response.get("isLast"),
    }
    return add_pagination_to_response(response, projects, limit, offset)


@tool(requires_auth=Atlassian(scopes=["read:jira-work"]))
async def get_project_by_id(
    context: ToolContext,
    project: Annotated[str, "The ID or key of the project to retrieve"],
) -> Annotated[dict[str, Any], "Information about the project"]:
    """Get the details of a Jira project by its ID or key."""
    client = JiraClient(context.get_auth_token_or_empty())

    try:
        response = await client.get(f"project/{project}")
    except NotFoundError:
        return {"error": f"Project not found: {project}"}

    cloud_name = cache.get_cloud_name(context.get_auth_token_or_empty())
    return {"project": clean_project_dict(response, cloud_name)}



import re
from datetime import datetime
from typing import Any

import pytz
from arcade_tdk.errors import ToolExecutionError
from kiota_abstractions.base_request_configuration import RequestConfiguration
from kiota_abstractions.headers_collection import HeadersCollection
from msgraph import GraphServiceClient
from msgraph.generated.users.item.mailbox_settings.mailbox_settings_request_builder import (
    MailboxSettingsRequestBuilder,
)

from arcade_microsoft.outlook_calendar.constants import WINDOWS_TO_IANA


def validate_date_times(start_date_time: str, end_date_time: str) -> None:
    """
    Validate date times are in ISO 8601 format and
    that end time is after start time (ignoring timezone offsets).

    Args:
        start_date_time: The start date time string to validate.
        end_date_time: The end date time string to validate.

    Raises:
        ValueError: If the date times are not in ISO 8601 format
        ToolExecutionError: If end time is not after start time.

    Note:
        This function ignores timezone offsets.
    """
    # parse into offset-aware datetimes
    start_aware = datetime.fromisoformat(start_date_time)
    end_aware = datetime.fromisoformat(end_date_time)

    # drop tzinfo to treat both as naïve local times
    start_naive = start_aware.replace(tzinfo=None)
    end_naive = end_aware.replace(tzinfo=None)

    if start_naive >= end_naive:
        raise ToolExecutionError(
            message="Start time must be before end time",
            developer_message=(
                f"The start time '{start_naive}' is not before the end time '{end_naive}'"
            ),
        )


def prepare_meeting_body(
    body: str, custom_meeting_url: str | None, is_online_meeting: bool
) -> tuple[str, bool]:
    """Prepare meeting body and determine final online meeting status.

    Args:
        body: The original meeting body text
        custom_meeting_url: Custom URL for the meeting, if one exists
        is_online_meeting: Whether this should be an online meeting

    Returns:
        tuple: (Updated meeting body, final online meeting status)

    Note:
        If a custom meeting URL is provided, is_online_meeting will be set to False
        to prevent Microsoft from generating its own meeting URL. The custom meeting
        URL will then be added to the body of the meeting.
    """
    is_online_meeting = not custom_meeting_url and is_online_meeting

    if custom_meeting_url:
        body = f"""{body}\n
.........................................................................
Join online meeting
{custom_meeting_url}"""

    return body, is_online_meeting


def validate_emails(emails: list[str]) -> None:
    """Validate a list of email addresses.

    Args:
        emails: The list of email addresses to validate.

    Raises:
        ToolExecutionError: If any email address is invalid.
    """
    invalid_emails = []
    for email in emails:
        if not is_valid_email(email):
            invalid_emails.append(email)
    if invalid_emails:
        raise ToolExecutionError(message=f"Invalid email address(es): {', '.join(invalid_emails)}")


def is_valid_email(email: str) -> bool:
    """Simple check to see if an email address is valid.

    Args:
        email: The email address to check.

    Returns:
        True if the email address is valid, False otherwise.
    """
    pattern = r"^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$"
    return re.match(pattern, email) is not None


def remove_timezone_offset(date_time: str) -> str:
    """Remove the timezone offset from the date_time string."""
    return re.sub(r"[+-][0-9]{2}:[0-9]{2}$|Z$", "", date_time)


def replace_timezone_offset(date_time: str, time_zone_offset: str) -> str:
    """Replace the timezone offset in the date_time string with the time_zone_offset.

    If the date_time str already contains a timezone offset, it will be replaced.
    If the date_time str does not contain a timezone offset, the time_zone_offset will be appended

    Args:
        date_time: The date_time string to replace the timezone offset in.
        time_zone_offset: The timezone offset to replace the existing timezone offset with.

    Returns:
        The date_time string with the timezone offset replaced or appended.
    """
    date_time = remove_timezone_offset(date_time)
    return f"{date_time}{time_zone_offset}"


def convert_timezone_to_offset(time_zone: str) -> str:
    """
    Convert a timezone (Windows or IANA) to ISO 8601 offset.
    First tries Windows timezone format, then IANA, then falls back to UTC if both fail.

    Args:
        time_zone: The timezone (Windows or IANA) to convert to ISO 8601 offset.

    Returns:
        The timezone offset in ISO 8601 format (e.g. '+08:00', '-07:00', or 'Z' for UTC)
    """
    # Try Windows timezone format
    iana_timezone = WINDOWS_TO_IANA.get(time_zone)
    if iana_timezone:
        try:
            tz = pytz.timezone(iana_timezone)
            now = datetime.now(tz)
            tz_offset = now.strftime("%z")

            if len(tz_offset) == 5:  # +HHMM format
                tz_offset = f"{tz_offset[:3]}:{tz_offset[3:]}"  # +HH:MM format
            return tz_offset  # noqa: TRY300
        except (pytz.exceptions.UnknownTimeZoneError, ValueError):
            pass

    # Try IANA timezone format
    try:
        tz = pytz.timezone(time_zone)
        now = datetime.now(tz)
        tz_offset = now.strftime("%z")

        if len(tz_offset) == 5:  # +HHMM format
            tz_offset = f"{tz_offset[:3]}:{tz_offset[3:]}"  # +HH:MM format
        return tz_offset  # noqa: TRY300
    except (pytz.exceptions.UnknownTimeZoneError, ValueError):
        # Fallback to UTC
        return "Z"


async def get_default_calendar_timezone(client: GraphServiceClient) -> str:
    """Get the authenticated user's default calendar's timezone.

    Args:
        client: The GraphServiceClient to use to get
            the authenticated user's default calendar's timezone.

    Returns:
        The timezone in "Windows timezone format" or "IANA timezone format".
    """
    query_params = MailboxSettingsRequestBuilder.MailboxSettingsRequestBuilderGetQueryParameters(
        select=["timeZone"]
    )
    request_config = RequestConfiguration(
        query_parameters=query_params,
    )
    response = await client.me.mailbox_settings.get(request_config)

    if response and response.time_zone:
        return response.time_zone
    return "UTC"


def create_timezone_headers(time_zone: str) -> HeadersCollection:
    """
    Create headers with timezone preference.

    Args:
        time_zone: The timezone to set in the headers.

    Returns:
        Headers collection with timezone preference set.
    """
    headers = HeadersCollection()
    headers.try_add("Prefer", f'outlook.timezone="{time_zone}"')
    return headers


def create_timezone_request_config(
    time_zone: str, query_parameters: Any | None = None
) -> RequestConfiguration:
    """
    Create a request configuration with timezone headers and optional query parameters.

    Args:
        time_zone: The timezone to set in the headers.
        query_parameters: Optional query parameters to include in the configuration.

    Returns:
        Request configuration with timezone headers and optional query parameters.
    """
    headers = create_timezone_headers(time_zone)
    return RequestConfiguration(
        headers=headers,
        query_parameters=query_parameters,
    )



from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    SimilarityCritic,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_github
from arcade_github.tools.issues import (
    create_issue,
    create_issue_comment,
)

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.9,
    warn_threshold=0.95,
)

catalog = ToolCatalog()
# Register the GitHub tools
catalog.add_module(arcade_github)


@tool_eval()
def github_issues_eval_suite() -> EvalSuite:
    """Evaluation suite for GitHub Issues tools."""
    suite = EvalSuite(
        name="GitHub Issues Tools Evaluation Suite",
        system_message="You are an AI assistant that helps users interact with GitHub issues using the provided tools.",
        catalog=catalog,
        rubric=rubric,
    )

    # Create Issue
    suite.add_case(
        name="Create a new issue",
        user_message="Create a new issue in the 'ArcadeAI/arcade-ai' repository with the title 'Bug: Login not working' and description 'Users are unable to log in to the application.' Assign the issue to TestUser, add it to milestone 1, and add the labels 'bug', and 'critical'.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=create_issue,
                args={
                    "owner": "ArcadeAI",
                    "repo": "arcade-ai",
                    "title": "Bug: Login not working",
                    "body": "Users are unable to log in to the application.",
                    "assignees": ["TestUser"],
                    "milestone": 1,
                    "labels": ["bug", "critical"],
                    "include_extra_data": False,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="owner", weight=0.2),
            BinaryCritic(critic_field="repo", weight=0.2),
            SimilarityCritic(critic_field="title", weight=0.2),
            SimilarityCritic(critic_field="body", weight=0.1),
            BinaryCritic(critic_field="assignees", weight=0.1),
            BinaryCritic(critic_field="milestone", weight=0.1),
            BinaryCritic(critic_field="labels", weight=0.1),
        ],
    )

    # Create Issue Comment
    suite.add_case(
        name="Add a comment to an existing issue",
        user_message="Add a comment to issue #42 in the 'ArcadeAI/test' repository saying 'This issue is being investigated by the dev team.'",
        expected_tool_calls=[
            ExpectedToolCall(
                func=create_issue_comment,
                args={
                    "owner": "ArcadeAI",
                    "repo": "test",
                    "issue_number": 42,
                    "body": "This issue is being investigated by the dev team.",
                    "include_extra_data": False,
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="owner", weight=0.2),
            SimilarityCritic(critic_field="repo", weight=0.2),
            BinaryCritic(critic_field="issue_number", weight=0.3),
            SimilarityCritic(critic_field="body", weight=0.2),
        ],
    )

    return suite



import httpx
import typer
from arcadepy import Arcade, NotFoundError
from rich.console import Console
from rich.table import Table

from arcade_cli.constants import (
    PROD_CLOUD_HOST,
    PROD_ENGINE_HOST,
)
from arcade_cli.utils import (
    OrderCommands,
    compute_base_url,
    validate_and_get_config,
)

console = Console()


app = typer.Typer(
    cls=OrderCommands,
    add_completion=False,
    no_args_is_help=True,
    pretty_exceptions_enable=False,
    pretty_exceptions_show_locals=False,
    pretty_exceptions_short=True,
)

state = {
    "engine_url": compute_base_url(
        host=PROD_ENGINE_HOST, port=None, force_tls=False, force_no_tls=False
    )
}


@app.callback()
def main(
    host: str = typer.Option(
        PROD_ENGINE_HOST,
        "--host",
        "-h",
        help="The Arcade Engine host.",
    ),
    port: int = typer.Option(
        None,
        "--port",
        "-p",
        help="The port of the Arcade Engine host.",
    ),
    force_tls: bool = typer.Option(
        False,
        "--tls",
        help="Whether to force TLS for the connection to the Arcade Engine.",
    ),
    force_no_tls: bool = typer.Option(
        False,
        "--no-tls",
        help="Whether to disable TLS for the connection to the Arcade Engine.",
    ),
) -> None:
    """
    Manage users in the system.
    """
    engine_url = compute_base_url(force_tls, force_no_tls, host, port)
    state["engine_url"] = engine_url


@app.command("list", help="List all workers")
def list_workers(
    cloud_host: str = typer.Option(
        PROD_CLOUD_HOST,
        "--cloud-host",
        "-c",
        help="The Arcade Engine host.",
        hidden=True,
    ),
    cloud_port: int = typer.Option(
        None,
        "--cloud-port",
        "-cp",
        help="The port of the Arcade Engine host.",
        hidden=True,
    ),
    force_tls: bool = typer.Option(
        False,
        "--tls",
        help="Whether to force TLS for the connection to the Arcade Engine.",
        hidden=True,
    ),
    force_no_tls: bool = typer.Option(
        False,
        "--no-tls",
        help="Whether to disable TLS for the connection to the Arcade Engine.",
        hidden=True,
    ),
) -> None:
    config = validate_and_get_config()
    engine_url = state["engine_url"]
    client = Arcade(api_key=config.api.key, base_url=engine_url)
    deployments = []
    try:
        cloud_url = compute_base_url(force_tls, force_no_tls, cloud_host, cloud_port)
        cloud_client = httpx.Client(base_url=cloud_url)
        response = cloud_client.get(
            "/api/v1/workers", headers={"Authorization": f"Bearer {config.api.key}"}
        )
        response.raise_for_status()
        deployments = response.json()["data"]["workers"]
    except Exception as e:
        console.log(f"Failed to get cloud deployments: {e}")

    print_worker_table(client, deployments)


def print_worker_table(client: Arcade, deployments: list[dict]) -> None:
    workers = client.workers.list()
    if not workers.items:
        console.print("No workers found", style="bold red")
        return

    # Create and print a table of worker information
    table = Table(title="Workers")
    table.add_column("ID")
    table.add_column("Cloud Deployed")
    table.add_column("Engine Registered")
    table.add_column("Enabled")
    table.add_column("Host")
    table.add_column("Toolkits")

    # Track workers that are registered in the engine
    engine_workers = []
    for worker in workers.items:
        if worker.id is None:
            continue
        engine_workers.append(worker.id)
        # Check if the worker is deployed in the cloud
        is_deployed = is_cloud_deployment(worker.id, deployments)
        # Get the toolkits for the worker

        tools = get_toolkits(client, worker.id)
        uri = worker.http.uri if worker.http and worker.http.uri else ""
        table.add_row(
            worker.id,
            str(is_deployed),
            str(True),
            str(worker.enabled),
            compare_endpoints(worker.id, uri, deployments),
            "Could not fetch toolkits" if tools == "" else tools,
        )
    for deployment in deployments:
        if deployment["name"] not in engine_workers:
            table.add_row(deployment["name"], "True", "False", "False", deployment["endpoint"], "")
    console.print(table)


# Check if the worker is in the list of cloud deployments
def is_cloud_deployment(name: str, deployments: list[dict]) -> bool:
    return any(deployment["name"] == name for deployment in deployments)


# Compare the endpoint of the worker in the engine to the endpoint in the cloud
# Return a highlighted diff if the endpoint in the engine is different from the endpoint in the cloud
def compare_endpoints(worker_id: str, engine_endpoint: str, deployments: list[dict]) -> str:
    if is_cloud_deployment(worker_id, deployments):
        for deployment in deployments:
            deployment_endpoint = deployment["endpoint"]
            if deployment_endpoint == engine_endpoint:
                return engine_endpoint
            return f"[red]Endpoint Mismatch[/red]\n[yellow]Registered Endpoint: {engine_endpoint}[/yellow]\n[green]Actual Endpoint:     {deployment_endpoint}[/green]"
    return engine_endpoint


@app.command("enable", help="Enable a worker")
def enable_worker(
    worker_id: str,
) -> None:
    config = validate_and_get_config()
    engine_url = state["engine_url"]
    arcade = Arcade(api_key=config.api.key, base_url=engine_url)
    try:
        arcade.workers.update(worker_id, enabled=True)
    except Exception as e:
        console.print(f"Error enabling worker: {e}", style="bold red")
        raise typer.Exit(code=1)


@app.command("disable", help="Disable a worker")
def disable_worker(
    worker_id: str,
) -> None:
    config = validate_and_get_config()
    engine_url = state["engine_url"]
    arcade = Arcade(api_key=config.api.key, base_url=engine_url)
    try:
        arcade.workers.update(worker_id, enabled=False)
    except Exception as e:
        console.print(f"Error disabling worker: {e}", style="bold red")
        raise typer.Exit(code=1)


@app.command("rm", help="Remove a worker")
def rm_worker(
    worker_id: str,
    engine_only: bool = typer.Option(
        False,
        "--deregister",
        "-d",
        help="Deregister the worker from the engine",
    ),
    cloud_host: str = typer.Option(
        PROD_CLOUD_HOST,
        "--cloud-host",
        "-c",
        help="The Arcade Engine host.",
        hidden=True,
    ),
    cloud_port: int = typer.Option(
        None,
        "--cloud-port",
        "-cp",
        help="The port of the Arcade Engine host.",
        hidden=True,
    ),
    force_tls: bool = typer.Option(
        False,
        "--tls",
        help="Whether to force TLS for the connection to the Arcade Engine.",
        hidden=True,
    ),
    force_no_tls: bool = typer.Option(
        False,
        "--no-tls",
        help="Whether to disable TLS for the connection to the Arcade Engine.",
        hidden=True,
    ),
) -> None:
    config = validate_and_get_config()
    engine_url = state["engine_url"]
    cloud_url = compute_base_url(force_tls, force_no_tls, cloud_host, cloud_port)

    # First attempt to delete from the cloud
    if not engine_only:
        try:
            client = httpx.Client()
            response = client.delete(
                f"{cloud_url}/api/v1/workers/{worker_id}",
                headers={"Authorization": f"Bearer {config.api.key}"},
            )
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 404:
                console.print(
                    "Deployment not found. To deregister the worker from the engine, use the --deregister flag.",
                    style="bold red",
                )
                raise typer.Exit(code=1)
            else:
                console.print(f"Error deleting deployment: {e}", style="bold red")
                raise typer.Exit(code=1)
        except Exception as e:
            console.print(f"Error deleting deployment: {e}", style="bold red")
            raise typer.Exit(code=1)

    # Then try to delete from the engine
    try:
        arcade = Arcade(api_key=config.api.key, base_url=engine_url)
        arcade.workers.delete(worker_id)
    except NotFoundError:
        console.print("Worker not found", style="bold red")
    except Exception as e:
        console.print(f"Error deleting worker from engine: {e}", style="bold red")
        raise typer.Exit(code=1)


@app.command("logs", help="Get logs for a worker")
def worker_logs(
    worker_id: str,
    cloud_host: str = typer.Option(
        PROD_CLOUD_HOST,
        "--cloud-host",
        "-c",
        help="The Arcade Engine host.",
        hidden=True,
    ),
    cloud_port: int = typer.Option(
        None,
        "--cloud-port",
        "-cp",
        help="The port of the Arcade Engine host.",
        hidden=True,
    ),
    force_tls: bool = typer.Option(
        False,
        "--tls",
        help="Whether to force TLS for the connection to the Arcade Engine.",
        hidden=True,
    ),
    force_no_tls: bool = typer.Option(
        False,
        "--no-tls",
        help="Whether to disable TLS for the connection to the Arcade Engine.",
        hidden=True,
    ),
) -> None:
    config = validate_and_get_config()
    cloud_url = compute_base_url(force_tls, force_no_tls, cloud_host, cloud_port)
    try:
        with httpx.stream(
            "GET",
            f"{cloud_url}/api/v1/workers/logs/{worker_id}",
            headers={"Authorization": f"Bearer {config.api.key}", "Accept": "text/event-stream"},
            # allow the connection to stay open indefinitely
            timeout=None,  # noqa: S113
        ) as s:
            for line in s.iter_lines():
                if not line or "[DONE]" in line:  # Skip empty lines
                    continue
                if "event: error" in line:
                    console.print("Could not stream logs", style="bold red")
                if line.startswith("data:"):
                    # Extract just the data portion after 'data:'
                    data = line[5:].strip()  # Remove 'data:' prefix and whitespace
                    console.print(data, markup=False)
    except Exception as e:
        console.print(f"Error connecting to log stream: {e}", style="bold red")
        raise typer.Exit(code=1)


def get_toolkits(client: Arcade, worker_id: str | None) -> str:
    if worker_id is None:
        return ""
    try:
        # Get tools for the given worker
        tools = client.workers.tools(worker_id)
        toolkits: list[str] = []
        if not tools.items:
            return ""

        # Get toolkit names
        for page in tools.iter_pages():
            for tool in page.items:
                if tool.toolkit.name not in toolkits:
                    toolkits.append(tool.toolkit.name)
        return ", ".join(toolkits)
    except NotFoundError:
        return ""
    except Exception as e:
        console.print(f"Error getting worker tools: {e}", style="bold red")
        raise typer.Exit(code=1)



import decimal
from decimal import Decimal
from typing import Annotated

from arcade_tdk import tool

decimal.getcontext().prec = 100


@tool
def add(
    a: Annotated[str, "The first number as a string"],
    b: Annotated[str, "The second number as a string"],
) -> Annotated[str, "The sum of the two numbers as a string"]:
    """
    Add two numbers together
    """
    # Use Decimal for arbitrary precision
    a_decimal = Decimal(a)
    b_decimal = Decimal(b)
    return str(a_decimal + b_decimal)


@tool
def subtract(
    a: Annotated[str, "The first number as a string"],
    b: Annotated[str, "The second number as a string"],
) -> Annotated[str, "The difference of the two numbers as a string"]:
    """
    Subtract two numbers
    """
    # Use Decimal for arbitrary precision
    a_decimal = Decimal(a)
    b_decimal = Decimal(b)
    return str(a_decimal - b_decimal)


@tool
def multiply(
    a: Annotated[str, "The first number as a string"],
    b: Annotated[str, "The second number as a string"],
) -> Annotated[str, "The product of the two numbers as a string"]:
    """
    Multiply two numbers together
    """
    # Use Decimal for arbitrary precision
    a_decimal = Decimal(a)
    b_decimal = Decimal(b)
    return str(a_decimal * b_decimal)


@tool
def divide(
    a: Annotated[str, "The first number as a string"],
    b: Annotated[str, "The second number as a string"],
) -> Annotated[str, "The quotient of the two numbers as a string"]:
    """
    Divide two numbers
    """
    # Use Decimal for arbitrary precision
    a_decimal = Decimal(a)
    b_decimal = Decimal(b)
    return str(a_decimal / b_decimal)


@tool
def sum_list(
    numbers: Annotated[list[str], "The list of numbers as strings"],
) -> Annotated[str, "The sum of the numbers in the list as a string"]:
    """
    Sum all numbers in a list
    """
    # Use Decimal for arbitrary precision
    return str(sum([Decimal(n) for n in numbers]))


@tool
def sum_range(
    start: Annotated[str, "The start of the range to sum as a string"],
    end: Annotated[str, "The end of the range to sum as a string"],
) -> Annotated[str, "The sum of the numbers in the list as a string"]:
    """
    Sum all numbers from start through end
    """
    return str(sum(list(range(int(start), int(end) + 1))))


@tool
def mod(
    a: Annotated[str, "The dividend as a string"],
    b: Annotated[str, "The divisor as a string"],
) -> Annotated[str, "The remainder after dividing a by b as a string"]:
    """
    Calculate the remainder (modulus) of one number divided by another
    """
    # Use Decimal for arbitrary precision
    return str(Decimal(a) % Decimal(b))



import decimal
from decimal import Decimal
from statistics import median as stats_median
from typing import Annotated

from arcade_tdk import tool

decimal.getcontext().prec = 100


@tool
def avg(
    numbers: Annotated[list[str], "The list of numbers as strings"],
) -> Annotated[str, "The average (mean) of the numbers in the list as a string"]:
    """
    Calculate the average (mean) of a list of numbers.
    Returns "0.0" if the list is empty.
    """
    # Use Decimal for arbitrary precision
    d_numbers = [Decimal(n) for n in numbers]
    return str(sum(d_numbers) / len(d_numbers)) if d_numbers else "0.0"


@tool
def median(
    numbers: Annotated[list[str], "A list of numbers as strings"],
) -> Annotated[str, "The median value of the numbers in the list as a string"]:
    """
    Calculate the median of a list of numbers.
    Returns "0.0" if the list is empty.
    """
    # Use Decimal for arbitrary precision
    d_numbers = [Decimal(n) for n in numbers]
    return str(stats_median(d_numbers)) if d_numbers else "0.0"



import asyncio
import json
import logging
import re
from collections.abc import Callable, Sequence
from datetime import datetime, timezone
from typing import Any, cast

from arcade_tdk import ToolContext
from arcade_tdk.errors import RetryableToolError

from arcade_slack.constants import (
    MAX_CONCURRENT_REQUESTS,
    MAX_PAGINATION_SIZE_LIMIT,
    MAX_PAGINATION_TIMEOUT_SECONDS,
)
from arcade_slack.custom_types import SlackPaginationNextCursor
from arcade_slack.exceptions import PaginationTimeoutError
from arcade_slack.models import (
    AbstractConcurrencySafeCoroutineCaller,
    BasicUserInfo,
    ConversationMetadata,
    ConversationType,
    ConversationTypeSlackName,
    Message,
    PaginationSentinel,
    SlackConversation,
    SlackConversationPurpose,
    SlackMessage,
    SlackUser,
    SlackUserList,
)

logger = logging.getLogger(__name__)


def format_users(user_list_response: SlackUserList) -> str:
    """Format a list of Slack users into a CSV string.

    Args:
        userListResponse: The response from the Slack API's users_list method.

    Returns:
        A CSV string with two columns: the users' name and real name, each user in a new line.
        The first line is the header with column names 'name' and 'real_name'.
    """
    csv_string = "name,real_name\n"
    for user in user_list_response["members"]:
        if not user.get("deleted", False):
            name = user.get("name", "")
            profile = user.get("profile", {})
            real_name = "" if not profile else profile.get("real_name", "")
            csv_string += f"{name},{real_name}\n"
    return csv_string.strip()


def format_conversations_as_csv(conversations: dict) -> str:
    """Format a list of Slack conversations into a CSV string.

    Args:
        conversations: The response from the Slack API's conversations_list method.

    Returns:
        A CSV string with the conversations' names.
    """
    csv_string = "All active Slack conversations:\n\nname\n"
    for conversation in conversations["channels"]:
        if not conversation.get("is_archived", False):
            name = conversation.get("name", "")
            csv_string += f"{name}\n"
    return csv_string.strip()


def remove_none_values(params: dict) -> dict:
    """Remove key/value pairs from a dictionary where the value is None.

    Args:
        params: The dictionary to remove None values from.

    Returns:
        A dictionary with None values removed.
    """
    return {k: v for k, v in params.items() if v is not None}


def get_slack_conversation_type_as_str(channel: SlackConversation) -> str | None:
    """Get the type of conversation from a Slack channel's dictionary.

    Args:
        channel: The Slack channel's dictionary.

    Returns:
        The type of conversation string in Slack naming standard.
    """
    if channel.get("is_channel"):
        return ConversationTypeSlackName.PUBLIC_CHANNEL.value
    if channel.get("is_group"):
        return ConversationTypeSlackName.PRIVATE_CHANNEL.value
    if channel.get("is_im"):
        return ConversationTypeSlackName.IM.value
    if channel.get("is_mpim"):
        return ConversationTypeSlackName.MPIM.value
    raise ValueError(f"Invalid conversation type in channel: {json.dumps(channel)}")


def convert_conversation_type_to_slack_name(
    conversation_type: ConversationType,
) -> ConversationTypeSlackName:
    """Convert a conversation type to another using Slack naming standard.

    Args:
        conversation_type: The conversation type enum value.

    Returns:
        The corresponding conversation type enum value using Slack naming standard.
    """
    mapping = {
        ConversationType.PUBLIC_CHANNEL: ConversationTypeSlackName.PUBLIC_CHANNEL,
        ConversationType.PRIVATE_CHANNEL: ConversationTypeSlackName.PRIVATE_CHANNEL,
        ConversationType.MULTI_PERSON_DIRECT_MESSAGE: ConversationTypeSlackName.MPIM,
        ConversationType.DIRECT_MESSAGE: ConversationTypeSlackName.IM,
    }
    return mapping[conversation_type]


def extract_conversation_metadata(conversation: SlackConversation) -> ConversationMetadata:
    """Extract conversation metadata from a Slack conversation object.

    Args:
        conversation: The Slack conversation dictionary.

    Returns:
        A dictionary with the conversation metadata.
    """
    conversation_type = get_slack_conversation_type_as_str(conversation)

    purpose: SlackConversationPurpose | None = conversation.get("purpose")
    purpose_value = "" if not purpose else purpose.get("value", "")

    metadata = ConversationMetadata(
        id=conversation.get("id"),
        name=conversation.get("name"),
        conversation_type=conversation_type,
        is_private=conversation.get("is_private", True),
        is_archived=conversation.get("is_archived", False),
        is_member=conversation.get("is_member", True),
        purpose=purpose_value,
        num_members=conversation.get("num_members", 0),
    )

    if conversation_type == ConversationTypeSlackName.IM.value:
        metadata["num_members"] = 2
        metadata["user"] = conversation.get("user")
        metadata["is_user_deleted"] = conversation.get("is_user_deleted")
    elif conversation_type == ConversationTypeSlackName.MPIM.value:
        conversation_name = conversation.get("name", "")
        if conversation_name:
            metadata["num_members"] = len(conversation_name.split("--"))
        else:
            metadata["num_members"] = None

    return metadata


def extract_basic_user_info(user_info: SlackUser) -> BasicUserInfo:
    """Extract a user's basic info from a Slack user dictionary.

    Args:
        user_info: The Slack user dictionary.

    Returns:
        A dictionary with the user's basic info.

    See https://api.slack.com/types/user for the structure of the user object.
    """
    profile = user_info.get("profile", {})
    display_name = None if not profile else profile.get("display_name")
    email = None if not profile else profile.get("email")
    return BasicUserInfo(
        id=user_info.get("id"),
        name=user_info.get("name"),
        is_bot=user_info.get("is_bot"),
        email=email,
        display_name=display_name,
        real_name=user_info.get("real_name"),
        timezone=user_info.get("tz"),
    )


def filter_conversations_by_user_ids(
    conversations: list[dict],
    user_ids: list[str],
    exact_match: bool = False,
) -> list[dict]:
    """
    Filter conversations by the members' user IDs.

    Args:
        conversations: The list of conversations to filter.
        user_ids: The user IDs to filter conversations for.
        exact_match: Whether to match the exact number of members in the conversations.

    Returns:
        The list of conversations found.
    """
    matches = []
    for conversation in conversations:
        member_ids = [member["id"] for member in conversation["members"]]
        if exact_match:
            same_length = len(user_ids) == len(member_ids)
            has_all_members = all(user_id in member_ids for user_id in user_ids)
            if same_length and has_all_members:
                matches.append(conversation)
        else:
            if all(user_id in member_ids for user_id in user_ids):
                matches.append(conversation)

    return matches


def is_user_a_bot(user: SlackUser) -> bool:
    """Check if a Slack user represents a bot.

    Args:
        user: The Slack user dictionary.

    Returns:
        True if the user is a bot, False otherwise.

    Bots are users with the "is_bot" flag set to true.
    USLACKBOT is the user object for the Slack bot itself and is a special case.

    See https://api.slack.com/types/user for the structure of the user object.
    """
    return user.get("is_bot") or user.get("id") == "USLACKBOT"


def is_user_deleted(user: SlackUser) -> bool:
    """Check if a Slack user represents a deleted user.

    Args:
        user: The Slack user dictionary.

    Returns:
        True if the user is deleted, False otherwise.

    See https://api.slack.com/types/user for the structure of the user object.
    """
    is_deleted = user.get("deleted")

    return is_deleted if isinstance(is_deleted, bool) else False


async def async_paginate(
    func: Callable,
    response_key: str | None = None,
    limit: int | None = None,
    next_cursor: SlackPaginationNextCursor | None = None,
    max_pagination_timeout_seconds: int = MAX_PAGINATION_TIMEOUT_SECONDS,
    sentinel: PaginationSentinel | None = None,
    *args: Any,
    **kwargs: Any,
) -> tuple[list, SlackPaginationNextCursor | None]:
    """Paginate a Slack AsyncWebClient's method results.

    The purpose is to abstract the pagination work and make it easier for the LLM to retrieve the
    amount of items requested by the user, regardless of limits imposed by the Slack API. We still
    return the next cursor, if needed to paginate further.

    Args:
        func: The Slack AsyncWebClient's method to paginate.
        response_key: The key in the response dictionary to extract the items from (optional). If
            not provided, the entire response dictionary is used.
        limit: The maximum number of items to retrieve (defaults to Slack's suggested limit).
        next_cursor: The cursor to use for pagination (optional).
        max_pagination_timeout_seconds: The maximum timeout for the pagination loop (defaults to
            MAX_PAGINATION_TIMEOUT_SECONDS).
        sentinel: Control whether the pagination should continue after each iteration (optional).
            If provided, the pagination will stop when the sentinel function returns True.
        *args: Positional arguments to pass to the Slack method.
        **kwargs: Keyword arguments to pass to the Slack method.

    Returns:
        A tuple containing the list of items and the next cursor, if needed to paginate further.
    """
    results: list[Any] = []

    async def paginate_loop() -> list[Any]:
        nonlocal results, next_cursor
        should_continue = True

        """
        The slack_limit variable makes the Slack API return no more than the appropriate
        amount of items. The loop extends results with the items returned and continues
        iterating if it hasn't reached the limit, and Slack indicates there're more
        items to retrieve.
        """

        while should_continue:
            iteration_limit = limit - len(results) if limit else MAX_PAGINATION_SIZE_LIMIT
            slack_limit = min(iteration_limit, MAX_PAGINATION_SIZE_LIMIT)
            iteration_kwargs = {**kwargs, "limit": slack_limit, "cursor": next_cursor}
            response = await func(*args, **iteration_kwargs)

            try:
                result = dict(response.data) if not response_key else response[response_key]
                results.extend(result)
            except KeyError:
                raise ValueError(f"Response key {response_key} not found in Slack response")

            next_cursor = response.get("response_metadata", {}).get("next_cursor")

            if (
                (sentinel and sentinel(last_result=result))
                or (limit and len(results) >= limit)
                or not next_cursor
            ):
                should_continue = False

        return results

    try:
        results = await asyncio.wait_for(paginate_loop(), timeout=max_pagination_timeout_seconds)
    # asyncio.TimeoutError for Python <= 3.10, TimeoutError for Python >= 3.11
    except (TimeoutError, asyncio.TimeoutError):
        raise PaginationTimeoutError(max_pagination_timeout_seconds)
    else:
        return results, next_cursor


def enrich_message_datetime(message: SlackMessage) -> Message:
    """Enrich message metadata with formatted datetime.

    It helps LLMs when they need to display the date/time in human-readable format. Slack
    will only return a unix-formatted timestamp (it's not actually UTC Unix timestamp, but
    the Unix timestamp in the user's timezone - I know, odd, but it is what it is).

    Args:
        message: The Slack message dictionary.

    Returns:
        The enriched message dictionary.
    """
    message = Message(**message)
    ts = message.get("ts")
    if isinstance(ts, str):
        try:
            timestamp = float(ts)
            message["datetime_timestamp"] = datetime.fromtimestamp(timestamp).strftime(
                "%Y-%m-%d %H:%M:%S"
            )
        except ValueError:
            pass
    return message


def convert_datetime_to_unix_timestamp(datetime_str: str) -> int:
    """Convert a datetime string to a unix timestamp.

    Args:
        datetime_str: The datetime string ('YYYY-MM-DD HH:MM:SS') to convert to a unix timestamp.

    Returns:
        The unix timestamp integer.
    """
    try:
        dt = datetime.strptime(datetime_str, "%Y-%m-%d %H:%M:%S")
        return int(dt.timestamp())
    except ValueError:
        raise RetryableToolError(
            "Invalid datetime format",
            developer_message=f"The datetime '{datetime_str}' is invalid. "
            "Please provide a datetime string in the format 'YYYY-MM-DD HH:MM:SS'.",
            retry_after_ms=500,
        )


def convert_relative_datetime_to_unix_timestamp(
    relative_datetime: str,
    current_unix_timestamp: int | None = None,
) -> int:
    """Convert a relative datetime string in the format 'DD:HH:MM' to unix timestamp.

    Args:
        relative_datetime: The relative datetime string ('DD:HH:MM') to convert to a unix timestamp.
        current_unix_timestamp: The current unix timestamp (optional). If not provided, the
            current unix timestamp from datetime.now is used.

    Returns:
        The unix timestamp integer.
    """
    if not current_unix_timestamp:
        current_unix_timestamp = int(datetime.now(timezone.utc).timestamp())

    days, hours, minutes = map(int, relative_datetime.split(":"))
    seconds = days * 86400 + hours * 3600 + minutes * 60
    return int(current_unix_timestamp - seconds)


def short_user_info(user: dict) -> dict[str, str | None]:
    data = {"id": user.get("id")}
    if user.get("name"):
        data["name"] = user["name"]
    if isinstance(user.get("profile"), dict) and user["profile"].get("email"):
        data["email"] = user["profile"]["email"]
    elif user.get("email"):
        data["email"] = user["email"]
    return data


def short_human_users_info(users: list[dict]) -> list[dict[str, str | None]]:
    return [short_user_info(user) for user in users if not user.get("is_bot")]


def is_valid_email(email: str) -> bool:
    """Validate an email address using regex.

    Args:
        email: The email address to validate.

    Returns:
        True if the email is valid, False otherwise.
    """
    email_pattern = r"^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$"
    return bool(re.match(email_pattern, email))


async def build_multiple_users_retrieval_response(
    context: ToolContext,
    users_responses: list[dict[str, Any]],
) -> list[dict[str, Any]]:
    """Builds response list for the get_multiple_users_by_usernames_or_emails function."""
    await raise_for_users_not_found(context, users_responses)

    users = []

    for users_response in users_responses:
        users.extend(users_response["users"])

    return cast(list[dict[str, Any]], users)


async def raise_for_users_not_found(
    context: ToolContext, users_responses: list[dict[str, Any]]
) -> None:
    """Raise an error if any user was not found in the responses."""
    users_not_found, available_users = collect_users_not_found_in_responses(users_responses)

    if users_not_found:
        not_found_message = ", ".join(users_not_found)
        s = "" if len(users_not_found) == 1 else "s"
        message = f"User{s} not found: {not_found_message}"
        available_users_prompt = await get_available_users_prompt(context, available_users)

        raise RetryableToolError(
            message=message,
            developer_message=message,
            additional_prompt_content=available_users_prompt,
            retry_after_ms=500,
        )


def collect_users_not_found_in_responses(
    responses: list[dict[str, Any]],
) -> tuple[list[str], list[dict[str, Any]]]:
    users_not_found = []
    available_users = []

    for response in responses:
        if response.get("not_found"):
            users_not_found.extend(response["not_found"])
        if response.get("available_users"):
            available_users = response["available_users"]

    return users_not_found, available_users


async def get_available_users_prompt(
    context: ToolContext,
    available_users: list[dict] | None = None,
    limit: int = 100,
) -> str:
    try:
        from arcade_slack.tools.users import list_users  # Avoid circular import

        if isinstance(available_users, list) and available_users:
            available_users = [
                user for user in available_users if not is_user_a_bot(SlackUser(**user))
            ]
            available_users_str = json.dumps(short_human_users_info(available_users))
            next_cursor = None
            potentially_more_users = True
        else:
            users = await list_users(context, limit=limit, exclude_bots=True)
            next_cursor = users["next_cursor"]
            available_users_str = json.dumps(short_human_users_info(users["users"]))
            potentially_more_users = bool(next_cursor)

        if not potentially_more_users:
            return f"The users available are: {available_users_str}"
        else:
            msg = (
                f"Some of the available users are: {available_users_str}. Potentially more users "
                f"can be retrieved by calling the 'Slack.{list_users.__tool_name__}' tool"
            )
            if next_cursor:
                msg += f" using the next cursor: '{next_cursor}' to continue pagination."
            return msg
    except Exception as e:
        return (
            "The tool tried to retrieve a list of available users, but failed with error: "
            f"{type(e).__name__}: {e!s}. Use the 'Slack.{list_users.__tool_name__}' tool "
            "to get a list of users."
        )


async def gather_with_concurrency_limit(
    coroutine_callers: Sequence[AbstractConcurrencySafeCoroutineCaller],
    semaphore: asyncio.Semaphore | None = None,
    max_concurrent_requests: int = MAX_CONCURRENT_REQUESTS,
) -> list[Any]:
    if not semaphore:
        semaphore = asyncio.Semaphore(max_concurrent_requests)

    return await asyncio.gather(*[caller(semaphore) for caller in coroutine_callers])  # type: ignore[no-any-return]


def cast_user_dict(user: dict[str, Any]) -> dict[str, Any]:
    slack_user = SlackUser(**cast(dict, user))
    return dict(**extract_basic_user_info(slack_user))


async def populate_users_in_messages(auth_token: str, messages: list[dict]) -> list[dict]:
    if not messages:
        return messages

    users = await get_users_from_messages(auth_token, messages)
    users_by_id = {user["id"]: {"id": user["id"], "name": user["name"]} for user in users}

    for message in messages:
        try:
            if "user" not in message or message.get("type") != "message":
                continue

            # Message author
            message["user"] = users_by_id.get(
                message.get("user"), {"id": message["user"], "name": None}
            )

            # User mentions in the message text
            text_mentions = re.findall(r"<@([A-Z0-9]+)>", message.get("text", ""))
            for user_id in text_mentions:
                if user_id in users_by_id:
                    user = users_by_id.get(user_id, {"id": user_id, "name": None})
                    name = user.get("name")
                    message["text"] = message["text"].replace(
                        f"<@{user_id}>", f"<@{name} (id:{user_id})>" if name else f"<@{user_id}>"
                    )

            # User mentions in reactions
            reactions = message.get("reactions")
            if isinstance(reactions, list):
                for reaction in reactions:
                    reaction_users = []
                    for user_id in reaction.get("users", []):
                        reaction_users.append(
                            users_by_id.get(user_id, {"id": user_id, "name": None})
                        )
                    reaction["users"] = reaction_users
        # If any data is missing, just leave the message as it is
        except Exception as exc:
            logger.exception(exc)  # noqa: TRY401

    return messages


async def get_users_from_messages(auth_token: str, messages: list[dict]) -> list[dict[str, Any]]:
    if not messages:
        return []

    from arcade_slack.user_retrieval import get_users_by_id  # Avoid circular import

    user_ids = get_user_ids_from_messages(messages)
    response = await get_users_by_id(auth_token, user_ids)

    return response["users"]


def get_user_ids_from_messages(messages: list[dict]) -> list[str]:
    if not messages:
        return []

    user_ids = []

    for message in messages:
        if message.get("type") != "message":
            continue

        # Message author
        user = message.get("user")
        if isinstance(user, str) and user:
            user_ids.append(user)

        # User mentions in the message text
        text = message.get("text")
        if isinstance(text, str) and text:
            user_ids.extend(re.findall(r"<@([A-Z0-9]+)>", text))

        # User mentions in reactions
        reactions = message.get("reactions")
        if isinstance(reactions, list):
            for reaction in reactions:
                user_ids.extend(reaction.get("users", []))

    return user_ids



import pytest

from arcade_stripe.tools.stripe import (
    create_billing_portal_session,
    create_customer,
    create_invoice,
    create_invoice_item,
    create_payment_link,
    create_price,
    create_product,
    create_refund,
    finalize_invoice,
    list_customers,
    list_invoices,
    list_payment_intents,
    list_prices,
    list_products,
    retrieve_balance,
)


class DummyContext:
    def get_secret(self, key: str):
        return "test_secret_key"


class DummyStripeAPI:
    def __init__(self, secret_key, context):
        self.secret_key = secret_key

    def run(self, method_name, **params):
        return {"method": method_name, "params": params}


@pytest.mark.parametrize(
    ("current_tool", "params"),
    [
        (create_customer, {"name": "John Doe"}),
        (create_customer, {"name": "John Doe", "email": "john.doe@example.com"}),
        (list_customers, {}),
        (list_customers, {"limit": 10}),
        (list_customers, {"email": "john.doe@example.com"}),
        (list_customers, {"limit": 10, "email": "john.doe@example.com"}),
        (create_product, {"name": "Product 1"}),
        (create_product, {"name": "Product 1", "description": "Description 1"}),
        (list_products, {}),
        (list_products, {"limit": 10}),
        (create_price, {"product": "product_123", "unit_amount": 1000, "currency": "usd"}),
        (list_prices, {}),
        (list_prices, {"product": "product_123"}),
        (list_prices, {"limit": 10}),
        (list_prices, {"product": "product_123", "limit": 10}),
        (create_payment_link, {"price": "price_123", "quantity": 100}),
        (list_invoices, {}),
        (list_invoices, {"customer": "customer_123"}),
        (list_invoices, {"limit": 10}),
        (list_invoices, {"customer": "customer_123", "limit": 10}),
        (create_invoice, {"customer": "customer_123"}),
        (create_invoice, {"customer": "customer_123", "days_until_due": 30}),
        (
            create_invoice_item,
            {"customer": "customer_123", "price": "price_123", "invoice": "invoice_123"},
        ),
        (finalize_invoice, {"invoice": "invoice_123"}),
        (retrieve_balance, {}),
        (create_refund, {"payment_intent": "payment_intent_123"}),
        (create_refund, {"payment_intent": "payment_intent_123", "amount": 100}),
        (list_payment_intents, {}),
        (list_payment_intents, {"customer": "customer_123"}),
        (list_payment_intents, {"limit": 10}),
        (list_payment_intents, {"customer": "customer_123", "limit": 10}),
        (create_billing_portal_session, {"customer": "customer_123"}),
        (
            create_billing_portal_session,
            {"customer": "customer_123", "return_url": "https://example.com"},
        ),
    ],
)
def test_stripe_tools(monkeypatch, current_tool, params):
    monkeypatch.setattr("arcade_stripe.tools.stripe.StripeAPI", DummyStripeAPI)

    context = DummyContext()

    result = current_tool(context, **params)
    expected = {"method": current_tool.__name__, "params": params}

    assert result == expected



# A conversation where a user asks for the structure of a space,
# and an LLM calls a tool to get the structure.
get_space_hierarchy_1 = [
    {"role": "system", "content": "Today is 2025-05-16, Friday."},
    {"role": "user", "content": "get structure of 98308"},
    {
        "role": "assistant",
        "content": "",
        "tool_calls": [
            {
                "id": "call_exyK4LmJEHSDn1Xw5oVfS9Xx",
                "type": "function",
                "function": {
                    "name": "Confluence_GetSpaceHierarchy",
                    "arguments": '{"space_identifier":"98308"}',
                },
            }
        ],
    },
    {
        "role": "tool",
        "content": '{"98418":{"children":[{"children":[],"id":"4653060","parent_id":"98418","parent_type":"TODO","title":"Mydb","type":"database","url":"https://ericconfluence.atlassian.net/wiki/spaces/testspace/database/4653060"},{"children":[{"children":[],"id":"4685837","parent_id":"4653064","parent_type":"TODO","title":"10-03-98","type":"page","url":"https://ericconfluence.atlassian.net/wiki/spaces/testspace/pages/4685837/10-03-98"}],"id":"4653064","parent_id":"98418","parent_type":"TODO","title":"DailyNotes","type":"folder","url":null},{"children":[{"children":[],"id":"5242883","parent_id":"5275650","parent_type":"TODO","title":"To be or not tobe","type":"page","url":"https://ericconfluence.atlassian.net/wiki/spaces/testspace/pages/5242883/To+be+or+not+to+be"},{"children":[],"id":"5275653","parent_id":"5275650","parent_type":"TODO","title":"Two roadsdiverged","type":"page","url":"https://ericconfluence.atlassian.net/wiki/spaces/testspace/pages/edit-v2/5275653"},{"children":[],"id":"5242903","parent_id":"5275650","parent_type":"TODO","title":"Eternity","type":"page","url":"https://ericconfluence.atlassian.net/wiki/spaces/testspace/pages/5242903/Eternity"},{"children":[],"id":"5242913","parent_id":"5275650","parent_type":"TODO","title":"BlackHole","type":"page","url":"https://ericconfluence.atlassian.net/wiki/spaces/testspace/pages/5242913/Black+Hole"}],"id":"5275650","parent_id":"98418","parent_type":"TODO","title":"Poetry","type":"folder","url":null},{"children":[],"id":"5275676","parent_id":"98418","parent_type":"TODO","title":"Drawings","type":"whiteboard","url":"https://ericconfluence.atlassian.net/wiki/spaces/testspace/whiteboard/5275676"}],"id":"98418","parent_id":null,"parent_type":null,"title":"testspace Home","type":"page","url":"https://ericconfluence.atlassian.net/wiki/spaces/testspace/overview"}}',  # noqa: E501
        "tool_call_id": "call_exyK4LmJEHSDn1Xw5oVfS9Xx",
        "name": "Confluence_GetSpaceHierarchy",
    },
    {
        "role": "assistant",
        "content": 'Here is the structure for the space with the identifier 98308:\n\n- **testspace Home** [Page](https://ericconfluence.atlassian.net/wiki/spaces/testspace/overview)\n  - **My db** [Database](https://ericconfluence.atlassian.net/wiki/spaces/testspace/database/4653060)\n  - **Daily Notes** (Folder)\n    - [10-03-98](https://ericconfluence.atlassian.net/wiki/spaces/testspace/pages/4685837/10-03-98) (Page)\n  - **Poetry** (Folder)\n    - [To be or not to be](https://ericconfluence.atlassian.net/wiki/spaces/testspace/pages/5242883/To+be+or+not+to+be) (Page)\n    - [Two roads diverged](https://ericconfluence.atlassian.net/wiki/spaces/testspace/pages/edit-v2/5275653) (Page)\n    - [Eternity](https://ericconfluence.atlassian.net/wiki/spaces/testspace/pages/5242903/Eternity) (Page)\n    - [Black Hole](https://ericconfluence.atlassian.net/wiki/spaces/testspace/pages/5242913/Black+Hole) (Page)\n  - **Drawings** [Whiteboard](https://ericconfluence.atlassian.net/wiki/spaces/testspace/whiteboard/5275676)\n\nThis structure includes various elements such as pages, a database, folders, and a whiteboard within the "testspace" space.',  # noqa: E501
    },
]

# A conversation where a user asks the assistant to create a page with content
create_page_1 = [
    {
        "role": "user",
        "content": "make a new page within the testspace space under the 4653064 folder. Within it, make a list of all the TODOs for the day which are 1. Write a blog 2. Walk the dog",  # noqa: E501
    },
    {
        "role": "assistant",
        "content": "",
        "tool_calls": [
            {
                "id": "call_xuvOuXpLAmX57RO1zj5hqQ1r",
                "type": "function",
                "function": {
                    "name": "Confluence_CreatePage",
                    "arguments": '{"space_identifier":"testspace","title":"TODOs forToday","content":"<h1>TODOs for Today</h1><ol><li>Write a blog</li><li>Walk thedog</li></ol>","parent_id":"4653064"}',  # noqa: E501
                },
            }
        ],
    },
    {
        "role": "tool",
        "content": '{"page":{"authorId":"712020:93b309a8-e9af-4615-b289-07bf23e6f13f","body":{"storage":{"representation":"storage","value":"\\u003ch1\\u003eTODOs for Today\\u003c/h1\\u003e\\u003col\\u003e\\u003cli\\u003eWrite a blog\\u003c/li\\u003e\\u003cli\\u003eWalk the dog\\u003c/li\\u003e\\u003c/ol\\u003e"}},"createdAt":"2025-05-16T22:29:59.969Z","id":"5439489","lastOwnerId":null,"ownerId":"712020:93b309a8-e9af-4615-b289-07bf23e6f13f","parentId":"4653064","parentType":"folder","position":1675,"spaceId":"98308","status":"current","title":"TODOs for Today","url":"https://ericconfluence.atlassian.net/wiki/spaces/testspace/pages/5439489/TODOs+for+Today","version":{"authorId":"712020:93b309a8-e9af-4615-b289-07bf23e6f13f","createdAt":"2025-05-16T22:29:59.969Z","message":"","minorEdit":false,"ncsStepVersion":"1","number":1}}}',  # noqa: E501
        "tool_call_id": "call_xuvOuXpLAmX57RO1zj5hqQ1r",
        "name": "Confluence_CreatePage",
    },
    {
        "role": "assistant",
        "content": 'I have created a new page titled "TODOs for Today" in the "testspace" space, under the folder with ID 4653064. The list of your tasks for the day is included:\n\n1. Write a blog\n2. Walk the dog\n\nYou can view the page [here](https://ericconfluence.atlassian.net/wiki/spaces/testspace/pages/5439489/TODOs+for+Today).',  # noqa: E501
    },
]

# A conversation where a user asks the assistant to list one space.
list_spaces_1 = [
    {"role": "system", "content": "Today is 2025-05-19, Monday."},
    {"role": "user", "content": "get 1 space"},
    {
        "role": "assistant",
        "content": "",
        "tool_calls": [
            {
                "id": "call_ur97pVeXzdB7ECNrr0zQBJmE",
                "type": "function",
                "function": {"name": "Confluence_ListSpaces", "arguments": '{"limit":1}'},
            }
        ],
    },
    {
        "role": "tool",
        "content": '{"pagination_token":"eyJpZCI6MjMyNjc5NCwic3BhY2VTb3J0T3JkZXIiOnsiZmllbGQiOiJOQU1FIiwiZGlyZWN0aW9uIjoiQVNDRU5ESU5HIn0sInNwYWNlU29ydE9yZGVyVmFsdWUiOiJlcmljYXJjYWRlIn0=","spaces":[{"authorId":"712020:f889c42e-f53e-4500-a7c6-706c3ef10951","createdAt":"2025-05-12T21:52:24.230Z","currentActiveAlias":"~712020f889c42ef53e4500a7c6706c3ef10951","description":null,"homepageId":"2327032","icon":null,"id":"2326794","key":"~712020f889c42ef53e4500a7c6706c3ef10951","name":"ericarcade","spaceOwnerId":"712020:f889c42e-f53e-4500-a7c6-706c3ef10951","status":"current","type":"personal","url":"https://ericconfluence.atlassian.net/wiki/spaces/~712020f889c42ef53e4500a7c6706c3ef10951"}]}',
        "tool_call_id": "call_ur97pVeXzdB7ECNrr0zQBJmE",
        "name": "Confluence_ListSpaces",
    },
    {
        "role": "assistant",
        "content": "Here is one space from Confluence:\n\n- **Name**: ericarcade\n- **Key**: ~712020f889c42ef53e4500a7c6706c3ef10951\n- **Type**:Personal\n- **Status**: Current\n- **Author ID**: 712020:f889c42e-f53e-4500-a7c6-706c3ef10951\n- **Created At**: 2025-05-12\n- **Homepage ID**:2327032\n- **URL**: [ericarcade Space](https://ericconfluence.atlassian.net/wiki/spaces/~712020f889c42ef53e4500a7c6706c3ef10951)",  # noqa: E501
    },
]



from arcade_google_maps.tools.google_maps import (
    get_directions_between_addresses,
    get_directions_between_coordinates,
)

__all__ = [
    "get_directions_between_addresses",
    "get_directions_between_coordinates",
]
